from __future__ import annotations

from functools import partial

from core_10x.named_constant import Enum

import rio
import ui_10x.platform_interface as i
from ui_10x.rio.component_builder import Widget


class FindFlags(Enum):
    MATCH_EXACTLY = ()


class ListItem(Widget, i.ListItem):
    __slots__ = ('_list_widget',)
    s_component_class = rio.SimpleListItem
    args = dict(key=lambda kwargs: kwargs['text'])

    def _make_kwargs(self, **kwargs):
        kwargs = super()._make_kwargs(**kwargs)
        del kwargs['align_y']
        return kwargs

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._list_widget = None

    def row(self):
        if self._list_widget:
            return self._list_widget.row(self)

    def set_selected(self, selected: bool):
        return self._list_widget.set_selected(self, selected)


class ListWidget(Widget, i.ListWidget):
    __slots__ = ('_on_press',)
    s_component_class = rio.ListView
    s_default_kwargs = dict(selection_mode='single')
    s_unwrap_single_child = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['grow_y'] = True

    def add_items(self, items: list[i.ListItem | str]):
        for item in items:
            self.add_item(item)

    def clicked_connect(self, bound_method):
        self._on_press = self.callback(bound_method)

    def _handle_on_press(self, item):
        if self._on_press:
            self._on_press(item)

    def add_item(self, item: ListItem | str):
        item = ListItem(text=item) if isinstance(item, str) else item
        item['on_press'] = partial(self._handle_on_press, item)
        item._list_widget = self
        self._kwargs[self.s_children_attr].append(item)
        self.force_update()

    def clear(self):
        self._kwargs['children'] = []
        self._kwargs['selected_items'] = []
        self.force_update()

    def find_items(self, text, flags):
        assert flags == FindFlags.MATCH_EXACTLY, 'only MatchExactly supported'
        return [item for item in self._kwargs[self.s_children_attr] if text == item['text']]

    def row(self, item: ListItem) -> int:
        return self._kwargs[self.s_children_attr].index(item)

    def take_item(self, row: int):
        try:
            return self._kwargs[self.s_children_attr].pop(row)
        finally:
            self.force_update()

    def set_selected(self, item: ListItem, selected: bool):
        # TODO: this is called from on-click handler for list item, which is
        # inefficient and duplicates rio built-in functionality
        selected_items = self.setdefault('selected_items', [])
        key = item['key']
        old_selected = key in selected_items
        print(selected_items, selected, old_selected, key)

        if not old_selected and selected:
            if self['selection_mode'] == 'single':
                selected_items = []
            selected_items.append(key)

        if old_selected and not selected:
            selected_items.remove(key)
        print(selected_items)
        self['selected_items'] = selected_items

    def set_enabled(self, enabled: bool):
        self['selection_mode'] = 'single' if enabled else 'none'
        for item in self._kwargs[self.s_children_attr]:
            item['on_press'] = partial(self._handle_on_press, item) if enabled else None



if __name__ == '__main__':

    from core_10x.logger import LOG

    LOG.begin('test_logger', log_level = LOG.DETAILED)

    LOG('First message')
    LOG.MEDIUM(dict(a = 1, b = 'ccc'))
    LOG.DETAILED('Detailed message')

    for i in range(10):
        LOG.VERBOSE(f'i = {i}')

    LOG('Last message')
    LOG.end()



from __future__ import annotations

from typing import TYPE_CHECKING

from core_10x.attic.backbone.backbone_store import BackboneStore
from core_10x.traitable import T, Traitable, TraitDefinition

if TYPE_CHECKING:
    from core_10x.traitable import RC
    from core_10x.ts_store import TsStore


class BackboneTraitable(Traitable):
    @classmethod
    def build_trait_dir(cls) -> RC:
        for trait_def in cls.__dict__.values():
            if isinstance(trait_def, TraitDefinition):
                trait_def.flags_change(T.EVAL_ONCE)

        return Traitable.build_trait_dir()

    @classmethod
    def store(cls) -> TsStore:
        return BackboneStore.bb_store()


class VaultTraitable(BackboneTraitable):
    @classmethod
    def store(cls) -> TsStore:
        return BackboneStore.store(BackboneStore.VAULT_DB_NAME)



import secrets
import keyring

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.serialization import load_pem_private_key, load_pem_public_key

from py10x_kernel import OsUser
from core_10x.global_cache import cache
from core_10x.environment_variables import EnvVars

PUBLIC_EXP = 65537
KEY_SIZE = 2048
PASSWORD_SIZE = 24
ENCODING = 'utf-8'


class SecKeys:
    @classmethod
    def generate_password(cls, length = PASSWORD_SIZE) -> str:
        return secrets.token_urlsafe(length)

    @classmethod
    @cache
    def retrieve_master_password(cls, throw = True) -> str:
        username = OsUser.me.name()
        v_mp_key = EnvVars.var.master_password_key
        if not v_mp_key:
            raise AssertionError(f'MasterPassword key may not be empty ({EnvVars.var_name(v_mp_key)})')

        pwd = keyring.get_password(v_mp_key.value, username)
        if pwd is None and throw:
            raise OSError(f'MasterPassword for {username} is not found')

        return pwd

    @classmethod
    def change_master_password(cls, password: str, override = False):
        username = OsUser.me.name()
        if not override and cls.retrieve_master_password(throw = False):
            raise AssertionError(f'MasterPassword for {username} is already set')

        keyring.set_password(EnvVars.master_password_key, username, password)

    @classmethod
    @cache
    def check_vault_uri(cls, throw = True) -> str:
        v_vault_uri = EnvVars.var.vault_ts_store_uri
        if not v_vault_uri and throw:
            raise OSError(f"Vault URI is not specified in either '{EnvVars.var_name(v_vault_uri)}' or {EnvVars.var_name('main_ts_store_uri')})")

        return v_vault_uri.value

    @classmethod
    @cache
    def retrieve_vault_password(cls, vault_uri: str = None, throw = True) -> str:
        username = OsUser.me.name()
        if vault_uri is None:
            vault_uri = cls.check_vault_uri(throw = throw)
            if vault_uri is None:
                return None

        pwd = keyring.get_password(vault_uri, username)
        if pwd is None and throw:
            raise OSError(f'Vault password for {username} @ {vault_uri} is not found')

        return pwd

    @classmethod
    def change_vault_password(cls, password: str, vault_uri: str = None, override = False):
        username = OsUser.me.name()
        if vault_uri is None:
            vault_uri = cls.check_vault_uri()

        if not override and cls.retrieve_vault_password(throw = False):
            raise AssertionError(f'Password for {username} @ {vault_uri} is already set')

        keyring.set_password(vault_uri, username, password)

    @classmethod
    def generate_keys(cls, pwd = None) -> tuple:
        private_key = rsa.generate_private_key(public_exponent = PUBLIC_EXP, key_size = KEY_SIZE, backend = default_backend())
        public_key = private_key.public_key()

        if pwd:
            format = serialization.PrivateFormat.PKCS8
            algo = serialization.BestAvailableEncryption(bytes(pwd, encoding = ENCODING))
        else:
            format = serialization.PrivateFormat.TraditionalOpenSSL
            algo = serialization.NoEncryption()

        private_key_pem = private_key.private_bytes(encoding = serialization.Encoding.PEM, format = format,
                                                    encryption_algorithm = algo)
        public_key_pem = public_key.public_bytes(encoding = serialization.Encoding.PEM,
                                                 format = serialization.PublicFormat.SubjectPublicKeyInfo)

        return (private_key_pem, public_key_pem)

    @classmethod
    def encrypt(cls, message, public_key_pem: bytes) -> bytes:
        if type(message) is str:
            message = bytes(message, encoding = ENCODING)

        public_key = load_pem_public_key(public_key_pem)
        return public_key.encrypt(
            message,
            padding.OAEP(
                mgf = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm = hashes.SHA256(),
                label = None
            )
        )

    @classmethod
    def decrypt(cls, encrypted_message: bytes, private_key_pem: bytes, to_str = True):
        private_key = load_pem_private_key(private_key_pem, password = None)
        res = private_key.decrypt(
            encrypted_message,
            padding.OAEP(
                mgf = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm = hashes.SHA256(),
                label = None
            )
        )

        if to_str:
            res = res.decode(encoding = ENCODING)

        return res

    @classmethod
    def encrypt_private_key(cls, private_key_pem: bytes, password) -> bytes:
        if type(password) is str:
            password = bytes(password, encoding = ENCODING)

        private_key = load_pem_private_key(private_key_pem, password = None)
        return private_key.private_bytes(
            encoding = serialization.Encoding.PEM,
            format = serialization.PrivateFormat.PKCS8,
            encryption_algorithm = serialization.BestAvailableEncryption(password),
        )

    @classmethod
    def decrypt_private_key(cls, private_key_with_password, password) -> bytes:
        if type(password) is str:
            password = bytes(password, encoding = ENCODING)

        pk = load_pem_private_key(private_key_with_password, password = password)
        return pk.private_bytes(
            encoding = serialization.Encoding.PEM,
            format = serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm = serialization.NoEncryption(),
        )

    def __init__(self, private_key_with_password: bytes, public_key_pem: bytes, password):
        if type(password) is str:
            password = bytes(password, encoding = ENCODING)

        self.private_key = load_pem_private_key(private_key_with_password, password = password)
        self.public_key = load_pem_public_key(public_key_pem)

    def encrypt_text(self, text: str) -> bytes:
        message = bytes(text, encoding = ENCODING)
        return self.public_key.encrypt(
            message,
            padding.OAEP(
                mgf = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm = hashes.SHA256(),
                label = None
            )
        )

    def decrypt_text(self, encrypted_message: bytes) -> str:
        res = self.private_key.decrypt(
            encrypted_message,
            padding.OAEP(
                mgf = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm = hashes.SHA256(),
                label = None
            )
        )
        return res.decode(encoding = ENCODING)



import keyring

from py10x_kernel import OsUser

from core_10x.traitable import Traitable, T, RT, RC, RC_TRUE, TsStore, cache
from core_10x.environment_variables import EnvVars


class VaultTraitable(Traitable):
    @classmethod
    @cache
    def retrieve_master_password(cls) -> str:
        username = OsUser.me.name()
        pwd = keyring.get_password(EnvVars.master_password_key, username)
        if pwd is None:
            raise OSError(f'XX MasterPassword for {username} is not found')

        return pwd

    @classmethod
    def keep_master_password(cls, password: str):
        username = OsUser.me.name()
        keyring.set_password(EnvVars.master_password_key, username, password)

    @classmethod
    @cache
    def retrieve_vault_password(cls, vault_uri: str) -> str:
        username = OsUser.me.name()
        pwd = keyring.get_password(vault_uri, username)
        if pwd is None:
            raise OSError(f'Password for {username} @ {vault_uri} is not found')

        return pwd

    @classmethod
    def keep_vault_password(cls, vault_uri: str, password: str):
        username = OsUser.me.name()
        keyring.set_password(vault_uri, username, password)

    @classmethod
    @cache
    def store_per_class(cls) -> TsStore:
        uri = EnvVars.var.vault_ts_store_uri.check()
        spec = TsStore.spec_from_uri(uri)
        kwargs = spec.kwargs
        ts_class = spec.resource_class
        hostname = kwargs[ts_class.HOSTNAME_TAG]
        is_running, with_auth = ts_class.is_running_with_auth(hostname)
        if not is_running or not with_auth:
            raise OSError(f'Vault host {hostname} must be run with auth')

        username = OsUser.me.name()
        kwargs[ts_class.USERNAME_TAG] = username
        kwargs[ts_class.PASSWORD_TAG] = cls.retrieve_master_password()

        return ts_class.instance(**kwargs)



import asyncio

import pytest
import rio.testing
from ui_10x import platform_interface as i
from ui_10x.rio.component_builder import DynamicComponent
from ui_10x.rio.widgets import LineEdit


async def test_line_edit_comprehensive() -> None:
    """Test LineEdit with comprehensive client-widget interaction verification."""
    widget = LineEdit('Initial Text')
    find_input = 'document.querySelector(".rio-input-box").querySelector("input")'
    find_tool_tip = 'document.querySelector(".rio-tooltip-popup").querySelector(".rio-text").children[0].innerText'

    edited_calls = []
    finished_calls = []

    def edited_handler(text):
        edited_calls.append(text)

    def finished_handler():
        finished_calls.append(True)

    widget.text_edited_connect(edited_handler)
    widget.editing_finished_connect(finished_handler)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # 1) Verify client shows widget value
        assert widget.text() == 'Initial Text'
        assert widget.text() == await test_client.execute_js(find_input + '.value')

        # 2) Modify client value (user typing)
        await test_client.execute_js(find_input + '.focus();')
        await test_client.execute_js(find_input + '.value = "User Typed Text";')
        await test_client.execute_js(find_input + '.dispatchEvent(new Event("input"));')
        await asyncio.sleep(1)  # wait for change delay

        # 3) Verify widget reflects client value
        assert widget.text() == 'User Typed Text'
        assert widget.text() == await test_client.execute_js(find_input + '.value')
        assert len(edited_calls) == 1
        assert widget.text() == edited_calls[0]

        # Test editing finished
        await test_client.execute_js(find_input + '.blur();')
        assert len(finished_calls) == 1
        assert widget.text() == await test_client.execute_js(find_input + '.value')

        # Test widget changes propagate to client with timeout protection
        widget.set_text('Widget Updated')
        await test_client.wait_for_refresh()
        client_value = await test_client.execute_js(find_input + '.value')
        assert client_value == 'Widget Updated'
        assert widget.text() == 'Widget Updated'

        # Test tooltip functionality
        widget.set_tool_tip('Helpful tip')
        await test_client.wait_for_refresh()
        await test_client._page.mouse.move(1, 1)
        tooltip_text = await test_client.execute_js(find_tool_tip)
        assert tooltip_text == 'Helpful tip'

        # Test password mode with timeout protection
        assert 'password' != await test_client.execute_js(find_input + '.type')
        widget.set_password_mode()
        await test_client.wait_for_refresh()
        assert 'password' == await test_client.execute_js(find_input + '.type')


async def test_line_edit_disabled_interaction() -> None:
    """Test LineEdit disabled state blocks user interaction."""

    right_button_presses = []
    left_button_presses = []

    class MyLineEdit(LineEdit):
        def mouse_press_event(self, event: i.MouseEvent):
            if event.is_right_button():
                right_button_presses.append(event)
            else:
                left_button_presses.append(event)

    widget = MyLineEdit('Initial Text')
    find_input = 'document.querySelector(".rio-input-box input")'

    edited_calls = []

    def edited_handler(text):
        edited_calls.append(text)

    widget.text_edited_connect(edited_handler)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial enabled state - typing should work
        async def test_interaction(old_content, new_content, enabled=True):
            initial_left = len(left_button_presses)
            initial_right = len(right_button_presses)
            initial_edited = len(edited_calls)
            await test_client.click(10, 20, button='right')  # handler
            await test_client._page.click('.rio-input-box input', click_count=3, force=True)  # select all
            await test_client._page.keyboard.press(new_content)
            await asyncio.sleep(1)

            right, left, edited, content = (1, 3, 1, new_content) if enabled else (1, 3, 0, old_content)

            assert len(right_button_presses) == right + initial_right
            assert len(left_button_presses) == left + initial_left
            assert len(edited_calls) == edited + initial_edited
            assert edited_calls[-1] == content
            assert widget.text() == content

        await test_interaction(widget.text(), 'A')

        # Disable widget
        widget.set_enabled(False)
        await test_client.wait_for_refresh()

        # Verify client shows disabled state
        client_disabled = await test_client.execute_js(find_input + '.disabled')
        assert client_disabled is True

        # Test that typing is blocked when disabled
        await test_interaction(widget.text(), 'B', enabled=False)

        # Re-enable widget
        widget.set_enabled(True)
        await test_client.wait_for_refresh()

        # Verify client shows enabled state
        client_disabled = await test_client.execute_js(find_input + '.disabled')
        assert client_disabled is False

        # Test that typing works again
        await test_interaction(widget.text(), 'C')



from core_10x.environment_variables import EnvVars
from core_10x.ts_store import TsStore
from core_10x.attic.vault.vault_traitable import VaultTraitable, T, RT, OsUser, cache
from core_10x.attic.vault.sec_keys import SecKeys

class VaultUser(VaultTraitable):
    # fmt: off
    user_id: str                    = T(T.ID)   // 'OS login'
    suspended: bool                 = T(False)

    private_key_encrypted: bytes    = T()
    public_key: bytes               = T()

    sec_keys: SecKeys               = RT(T.EVAL_ONCE)
    # fmt: on

    def user_id_get(self) -> str:
        return self.__class__.myname()

    def sec_keys_get(self) -> SecKeys:
        return SecKeys(self.private_key_encrypted, self.public_key, VaultTraitable.retrieve_master_password())

    @classmethod
    def create_new(cls, master_password: str, save = True) -> 'VaultUser':
        VaultTraitable.keep_master_password(master_password)
        private_key_pem, public_key_pem = SecKeys.generate_keys()
        me = VaultUser()
        me.public_key = public_key_pem
        me.private_key_encrypted = SecKeys.encrypt_private_key(private_key_pem, master_password)

        if save:
            me.save().throw()

        return me

    @classmethod
    def new_vault(cls, vault_uri: str, password: str, master_password: str = None):
        username = cls.myname()
        vault_db = TsStore.instance_from_uri(vault_uri, username = username, password = password)
        with vault_db:
            me = cls.existing_instance(user_id = username, _throw = False)
            if not me:
                if not master_password:
                    raise AssertionError('Master password is required')
                me = cls.create_new(master_password)
            else:
                try:
                    existing_master_password = cls.retrieve_master_password()
                except Exception:
                    existing_master_password = None
                if not existing_master_password:
                    raise AssertionError('No existing MasterPassword found')
                if existing_master_password != master_password:
                    raise AssertionError('MasterPassword provided does not match the stored one')

            cls.keep_vault_password(vault_uri, password)

    @classmethod
    def is_functional_account(cls, user_id: str) -> bool:
        return user_id.split('-', 1) == EnvVars.functional_account_prefix

    @classmethod
    @cache
    def myname(cls) -> str:
        return OsUser.me.name()

    @classmethod
    @cache
    def me(cls) -> 'VaultUser':
        return cls.existing_instance(user_id=cls.myname())



from __future__ import annotations

import abc
import inspect
from collections import deque
from urllib.parse import quote, urlencode, urlsplit, urlunsplit

#-- TODO: move to attic
# class ResourceRequirements:
#     def __init__(self, resource_type, *args, **kwargs):
#         self.domain = None
#         self.category: str = None
#         self.resource_type = resource_type
#         self.args = args
#         self.kwargs = kwargs
#
#
# class ResourceBinding:
#     def __init__(
#         self, _resource_class: type = None, _resource_type: ResourceType = None, _resource_name: str = None, _driver_name: str = None, **kwargs
#     ):
#         if _resource_class:
#             assert issubclass(_resource_class, Resource), f'{_resource_class} is not a subclass of Resource'
#         else:
#             if _resource_name:
#                 _resource_type = ResourceType.instance(_resource_name)
#             else:
#                 assert _resource_type and isinstance(_resource_type, ResourceType), '_resource_type must be an instance of ResourceType'
#
#             _resource_class = _resource_type.resource_drivers.get(_driver_name)
#             assert _resource_class, f'Unknown _driver_name {_driver_name}'
#
#         self.resource_class = _resource_class
#         self.kwargs = kwargs
#
#
# R = ResourceBinding
#

class ResourceType:
    s_dir = {}

    def __init__(self, name: str):
        xrt = self.s_dir.get(name)
        assert xrt is None, f"Resource type '{name}' has already been created"
        self.s_dir[name] = xrt

        self.name = name
        self.resource_stack = deque()
        self.resource_drivers = {}

    # def __call__(self, *args, **kwargs):
    #     return ResourceRequirements(self, *args, **kwargs)

    def __getattr__(self, driver_name: str):
        return self.resource_drivers.get(driver_name)

    @staticmethod
    def instance(name: str, throw: bool = True):
        rt = ResourceType.s_dir.get(name)
        if not rt and throw:
            raise ValueError(f"Unknown Resource type '{name}'")

        return rt

    def register_driver(self, name: str, driver_class):
        existing_driver = self.resource_drivers.get(name)
        assert not existing_driver, f"Resource '{name}' is already registered in {existing_driver}"
        assert inspect.isclass(driver_class) and issubclass(driver_class, Resource), 'driver_class must be a subclass of Resource'
        self.resource_drivers[name] = driver_class

    def resource_driver(self, name: str, throw: bool = True):
        r = self.resource_drivers.get(name)
        if not r and throw:
            raise ValueError(f"Unknown resource '{name}'")

        return r

    def begin_using(self, resource, last: bool = True):
        self.resource_stack.append(resource) if last else self.resource_stack.appendleft(resource)

    def end_using(self):
        self.resource_stack.pop()  # -- will throw if begin_using() hasn't been called

    def current_resource(self):
        stack = self.resource_stack
        return stack[-1] if stack else None


class ResourceSpec:
    def __init__(self, resource_class, kwargs: dict):
        self.resource_class = resource_class
        self.kwargs = kwargs

    def hostname(self) -> str:
        return self.kwargs[Resource.HOSTNAME_TAG]

    def set_credentials(self, username: str = None, password: str = None):
        if username is not None:
            self.kwargs[self.resource_class.USERNAME_TAG] = username
        if password is not None:
            self.kwargs[self.resource_class.PASSWORD_TAG] = password

    def uri(self) -> str:
        kwargs = dict(self.kwargs)
        protocol = kwargs.pop(Resource.PROTOCOL_TAG, None)
        if not protocol:
            raise ValueError('URI protocol is missing.')

        username = kwargs.pop(Resource.USERNAME_TAG, None)
        password = kwargs.pop(Resource.PASSWORD_TAG, None)
        userinfo = ''
        if username:
            userinfo = quote(username, safe='')
            if password:
                userinfo += f':{quote(password, safe="")}'
            userinfo += '@'

        host_netloc = kwargs.pop(Resource.NETLOC_TAG, None)
        if host_netloc is None:     #-- kwargs were constructed manually, not via parse_uri
            host = kwargs.pop(Resource.HOSTNAME_TAG, '')
            port = kwargs.pop(Resource.PORT_TAG, None)
            host_netloc = f'{host}:{port}' if port else host
        else:
            kwargs.pop(Resource.HOSTNAME_TAG, None)
            kwargs.pop(Resource.PORT_TAG, None)

        dbname   = kwargs.pop(Resource.DBNAME_TAG, None)
        query    = kwargs.pop(Resource.QUERY_TAG, '')
        fragment = kwargs.pop(Resource.FRAGMENT_TAG, '')
        if kwargs:
            extra = urlencode(kwargs, doseq=True)
            query = f'{query}&{extra}' if query else extra

        #-- always emit '//' so empty-netloc URIs like duckdb:// or duckdb:///path round-trip correctly;
        #-- urlunsplit drops '//' for unknown schemes when netloc is empty
        netloc = userinfo + host_netloc
        path   = f'/{dbname}' if dbname else ''
        url    = f'{protocol}://{netloc}{path}'
        if query:    url += f'?{query}'
        if fragment: url += f'#{fragment}'
        return url

class Resource(abc.ABC):
    PROTOCOL_TAG    = 'protocol'
    NETLOC_TAG      = 'netloc'      #-- host+port portion of netloc, as formatted by urlsplit (IPv6 brackets preserved)
    HOSTNAME_TAG    = 'hostname'
    PORT_TAG        = 'port'
    USERNAME_TAG    = 'username'
    DBNAME_TAG      = 'dbname'
    PASSWORD_TAG    = 'password'
    QUERY_TAG       = 'query'
    FRAGMENT_TAG    = 'fragment'
    SSL_TAG         = 'ssl'

    s_resource_type: ResourceType = None
    s_driver_name: str = None

    @classmethod
    def parse_uri(cls, uri: str) -> dict:
        try:
            parts = urlsplit(uri)
            if not parts.scheme:
                raise ValueError(f'Invalid URI = {uri}')

            kwargs = {
                cls.PROTOCOL_TAG: parts.scheme,
                cls.NETLOC_TAG:   parts.netloc.split('@', 1)[-1],   #-- strip userinfo; urlsplit already formatted host (IPv6 brackets etc.)
                cls.USERNAME_TAG: parts.username,
                cls.PASSWORD_TAG: parts.password,
            }

            if parts.hostname is not None:
                kwargs[cls.HOSTNAME_TAG] = parts.hostname
            if parts.port is not None:
                kwargs[cls.PORT_TAG] = parts.port

            db_name = parts.path.lstrip('/')
            if db_name:
                kwargs[cls.DBNAME_TAG] = db_name
            if parts.query:
                kwargs[cls.QUERY_TAG] = parts.query
            if parts.fragment:
                kwargs[cls.FRAGMENT_TAG] = parts.fragment

            return kwargs
        except Exception as e:
            raise ValueError(f'Invalid URI = {uri}') from e

    def __init_subclass__(cls, resource_type: ResourceType = None, resource_name: str = None, **kwargs):
        if cls.s_resource_type is None:     #-- must be a top class of a particular resource type, e.g. TsStore
            assert resource_type and isinstance(resource_type, ResourceType), 'instance of ResourceType is expected'
            assert resource_name is None, f'May not define Resource name for top class of Resource Type: {resource_type}'
            cls.s_resource_type = resource_type

        else:   #-- a Resource of a particular resource type
            assert resource_type is None, f'resource_type is already set: {cls.s_resource_type}'
            assert resource_name and isinstance(resource_name, str), 'a unique Resource name is expected'
            cls.s_resource_type.register_driver(resource_name, cls)
            cls.s_driver_name = resource_name

    def __enter__(self):
        return self.begin_using()

    def __exit__(self, *args):
        self.end_using()

    def begin_using(self):
        rt = self.__class__.s_resource_type
        rt.begin_using(self)
        self.on_enter()
        return rt

    def end_using(self):
        self.__class__.s_resource_type.end_using()
        self.on_exit()

    @classmethod
    def instance_from_uri(cls, uri: str, username: str = None, password: str = None) -> Resource:
        spec = cls.spec_from_uri(uri)
        spec.set_credentials(username=username, password=password)
        return spec.resource_class.instance(**spec.kwargs)

    @classmethod
    def spec_from_uri(cls, uri: str) -> ResourceSpec:
        return ResourceSpec(cls, cls.parse_uri(uri))

    @classmethod
    @abc.abstractmethod
    def instance(cls, *args, **kwargs) -> Resource: ...

    @abc.abstractmethod
    def on_enter(self): ...

    @abc.abstractmethod
    def on_exit(self): ...

class NullResource:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

NULL_RESOURCE = NullResource()

#=========== Known Resource Types
TS_STORE        = ResourceType('TS_STORE')
REL_DB          = ResourceType('REL_DB')
CLOUD_CLUSTER   = ResourceType('CLOUD_CLUSTER')



from __future__ import annotations

import rio


class LineEditComponent(rio.Component):
    text: str | None = None
    tooltip: str | None = None
    is_sensitive: bool = True
    on_change: rio.EventHandler[[str]] = None
    on_lose_focus: rio.EventHandler[[str]] = None
    text_style: rio.TextStyle | None = None
    is_secret: bool = False
    on_pointer_up: rio.EventHandler[[rio.PointerEvent]] = None

    def build(self):
        component = rio.TextInput(
            self.bind().text,
            is_sensitive=self.is_sensitive,
            on_change=self.on_change,
            on_lose_focus=self.on_lose_focus,
            text_style=self.text_style,
            is_secret=self.is_secret,
        )

        if self.on_pointer_up is not None:
            component = rio.PointerEventListener(
                component,
                on_pointer_up=self.on_pointer_up,
                consume_events=False,
                event_order='before-child',
            )

        if self.tooltip is not None:
            component = rio.Tooltip(anchor=component, tip=self.tooltip)

        return component



from __future__ import annotations

from core_10x.global_cache import cache

from ui_10x.rio import component_builder, widgets


@cache
def init(): ...


class Object: ...


Application = widgets.Application


DirectConnection = component_builder.ConnectionType.DIRECT
QueuedConnection = component_builder.ConnectionType.QUEUED


SignalDecl = component_builder.SignalDecl


def signal_decl(arg=object):
    assert arg is object, 'arg must be object'
    return SignalDecl()


MouseEvent = component_builder.MouseEvent


SCROLL = widgets.SCROLL

Point = component_builder.Point
FontMetrics = component_builder.FontMetrics
SizePolicy = component_builder.FontMetrics
TEXT_ALIGN = component_builder.TEXT_ALIGN

Widget = component_builder.Widget
Layout = component_builder.Layout
FlowLayout = component_builder.FlowLayout

LineEdit = widgets.LineEdit
Label = widgets.Label
PushButton = widgets.PushButton

Spacer = widgets.Spacer

HBoxLayout = widgets.HBoxLayout
VBoxLayout = widgets.VBoxLayout
FormLayout = widgets.FormLayout

Dialog = widgets.Dialog

MessageBox = widgets.MessageBox

RadioButton = widgets.RadioButton
ButtonGroup = widgets.ButtonGroup

GroupBox = widgets.GroupBox

TextEdit = widgets.TextEdit
CheckBox = widgets.CheckBox
ScrollArea = widgets.ScrollArea
Separator = widgets.Separator


def separator(horizontal=True) -> Separator:
    return Separator() if horizontal else Separator(orientation='vertical')


Direction = widgets.Direction

Vertical = Direction.VERTICAL
Horizontal = Direction.HORIZONTAL

Splitter = widgets.Splitter

Style = widgets.Style

FindFlags = widgets.FindFlags
MatchExactly = FindFlags.MATCH_EXACTLY

ListWidget = widgets.ListWidget
ListItem = widgets.ListItem

TreeWidget = widgets.TreeWidget
TreeItem = widgets.TreeItem

CalendarWidget = widgets.CalendarWidget



from __future__ import annotations

from ui_10x.rio.component_builder import UserSessionContext

import rio

theme = rio.Theme.from_colors(
    primary_color=rio.Color.from_hex('01dffdff'),
    secondary_color=rio.Color.from_hex('0083ffff'),
)


def on_session_start(session):
    session.attach(UserSessionContext(host='localhost', dbname='test'))  # TODO: backbone


app = rio.App(
    name='ui_10x.rio.apps.examples',
    description='Rio pages based on components created using ui_10x.platform_interface',
    theme=theme,
    on_session_start=on_session_start,
)



if __name__ == '__main__':
    from datetime import date, datetime

    from core_10x.named_constant import Enum, EnumBits
    from core_10x.nucleus import Nucleus

    class COLOR(Enum):
        RED     = ()
        GREEN   = ()
        BLUE    = ()

    class STATE(EnumBits):
        RUNNING     = ()
        SUSPENDED   = ()
        ABORTED     = ()

    data = {
        bool:           False,
        int:            100,
        type(None):     None,
        float:          1e-6,
        complex:        complex(5.43, -7.1),
        str:            'plain text',
        datetime:       datetime.now(),
        date:           date.today(),
        list:           [ 10, 5.1, 'label', datetime.now(), date(2020, 1, 1), ['a', 100], dict( a = 1, b = 2) ],
        tuple:          ( 10, 5.1, 'label', datetime.now(), date(2020, 1, 1), ['a', 100], dict( a = 1, b = 2) ),
        dict:           { 1: 1, 'a': 'abc', 'b': [ 10, 100. ], 'c': dict(x = -100, y = -200) },
        COLOR:          COLOR.GREEN,
        STATE:          STATE.RUNNING | STATE.SUSPENDED
    }

    serialized_data     = { dt: Nucleus.serialize_any(v, False) for dt, v in data.items() }
    deserialized_data   = { dt: Nucleus.deserialize_any(v, None) for dt, v in serialized_data.items() }

    for dt, value in data.items():
        s_value = deserialized_data.get(dt, deserialized_data)
        if s_value is deserialized_data:
            print(f'{dt} is missing')
        else:
            if s_value != value:
                print(f'{dt}:\nexpected\n{value}\ngot\n{s_value}')



from core_10x.traitable import RT, M, Trait, Traitable


class A(Traitable):
    name: str = RT()


class B(A):
    name: int = M()


def name_ui_flags(cls):
    t: Trait = cls.trait('name')
    return t.ui_hint.flags


def test_uihint_modification():
    flags = (name_ui_flags(A), name_ui_flags(B))
    assert flags[0] == flags[1]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from core_10x.rc import RC, RC_TRUE
from core_10x.trait import T, Ui

from ui_10x.utils import UxStyleSheet, ux

if TYPE_CHECKING:
    from core_10x.trait import Trait


class TraitWidget:
    # __slots__ = ('program_edit', 'refresh_context', 'style_sheet', 'trait', 'trait_editor')
    s_hinted_widgets = {}

    def __init_subclass__(cls, widget_type: Ui.WIDGET_TYPE = None, **kwargs):
        if widget_type:
            TraitWidget.s_hinted_widgets[widget_type] = cls

    @staticmethod
    def widget_class(widget_type: Ui.WIDGET_TYPE):
        w_cls = TraitWidget.s_hinted_widgets.get(widget_type)
        assert w_cls, f'Unknown trait widget type: {widget_type}'
        return w_cls

    @staticmethod
    def instance(trait_editor) -> TraitWidget | None:
        w_type = trait_editor.ui_hint.widget_type
        if w_type is not Ui.WIDGET_TYPE.NONE:
            w_cls = TraitWidget.widget_class(w_type)
            return w_cls(trait_editor)
        return None

    def __init__(self, trait_editor):
        self.trait_editor = trait_editor
        self.trait: Trait = trait_editor.trait
        self.program_edit = False

        self._create()

        if self.trait_editor.is_read_only():
            self._set_read_only(True)

        self.set_tool_tip(self.trait.ui_hint.tip)

        traitable = self.trait_editor.traitable

        self.style_sheet = sheet = UxStyleSheet(self)
        sh = traitable.get_style_sheet(self.trait)
        sheet.update(sh)

        value = traitable.get_value(self.trait)
        self.set_widget_value(value)

        self.refresh_context = ctx = RefreshContext(self)
        traitable.create_ui_node(self.trait, ctx.emit_signal)

    def refresh(self):
        traitable = self.trait_editor.traitable
        trait = self.trait
        traitable.update_ui_node(self.trait)

        if not trait.flags_on(T.EXPENSIVE):
            value = traitable.get_value(trait)
            self.set_widget_value(value)
        else:
            sh = traitable.get_style_sheet(trait)
            self.style_sheet.update(sh)

            if not traitable.is_valid(trait):
                self.style_sheet.update({Ui.BG_COLOR: 'lightblue'}, _system=True)

    def widget_value(self):
        return self._value()

    def set_widget_value(self, value) -> bool:
        self.program_edit = True
        try:
            self._set_value(value)
        except Exception:  # -- the real widget is gone (so far, has encountered this only in Qt on Mac OS)
            # self.clean()   #-- TODO: fix!
            self.program_edit = False
            return False

        sh = self.trait_editor.traitable.get_style_sheet(self.trait)
        self.style_sheet.update(sh)
        self.program_edit = False
        return True

    def update_trait_value(self, value: Any = None, invalidate: bool = False):
        if not self.program_edit:
            traitable = self.trait_editor.traitable
            if invalidate:
                traitable.invalidate_value(self.trait)
                rc = RC_TRUE
            else:
                if value is None:
                    value = self.widget_value()

                try:
                    rc = traitable.set_value(self.trait, value)
                except Exception as e:
                    rc = RC(False, str(e))

            if not rc:
                self.set_tool_tip(rc.error())
                self.style_sheet.update({Ui.BG_COLOR: 'orange'}, _system=True)
            else:
                self.set_tool_tip(self.trait.ui_hint.tip)
                self.style_sheet.restore()

    # fmt: off
    def _create(self):              raise NotImplementedError
    def _set_read_only(self, flag): raise NotImplementedError
    def _value(self):               raise NotImplementedError
    def _set_value(self, value):    raise NotImplementedError
    # fmt: on


class RefreshContext(ux.Object):
    REFRESH = ux.signal_decl(object)

    def __init__(self, trait_widget: TraitWidget):
        super().__init__()
        self.trait_widget = trait_widget
        self.REFRESH.connect(TraitWidget.refresh, type=ux.QueuedConnection)

    def emit_signal(self):
        self.REFRESH.emit(self.trait_widget)



from typing import Any

from core_10x.trait_filter import f
from core_10x.traitable import RC, RT, T, Traitable
from core_10x.traitable_id import ID

from ui_10x.entity_stocker import EntityStocker, StockerPlug
from ui_10x.traitable_editor import TraitableEditor
from ui_10x.utils import (
    UxSearchableList,
    ux,
    ux_make_scrollable,
    ux_push_button,
    ux_warning,
)


class Collection(Traitable):
    cls: type[Traitable]
    filter: f = RT(T.HIDDEN)

    entity_ids: list[str]

    def cls_set(self, t, cls) -> RC:
        assert issubclass(cls, Traitable), f'{cls} is not a Traitable class'
        return self.raw_set_value(t, cls)

    def entity_ids_get(self) -> list[str]:
        return [entity_id.value for entity_id in self.cls.load_ids()]

    def refresh(self):
        self.invalidate_value('entity_ids')


class CollectionEditor(Traitable):
    coll: Collection
    current_class: Any
    coll_title: str
    num_panes: int = RT(1)

    main_w: ux.Splitter
    searchable_list: UxSearchableList
    stocker: EntityStocker

    current_editor: Any
    current_entity: Traitable

    def current_class_get(self):
        return self.coll.cls

    def stocker_get(self):
        return EntityStocker(plug=StockerPlug(master=self))

    def main_widget(self) -> ux.Widget:
        sp = ux.Splitter(ux.Horizontal)

        left_w = ux.Widget()
        left_lay = ux.VBoxLayout()
        left_w.set_layout(left_lay)

        left_top_lay = ux.HBoxLayout()
        line_w = ux.LineEdit()
        line_w.set_minimum_width(100)
        button_w = ux_push_button('New', callback=self.on_new_entity, style_icon='FileIcon')
        left_top_lay.add_widget(line_w)
        left_top_lay.add_widget(button_w)

        left_lay.add_layout(left_top_lay)

        # fmt: off
        self.searchable_list = list_w = UxSearchableList(
            text_widget = line_w,
            title       = f'Instances of {self.coll.cls.__name__}',
            choices     = self.coll.entity_ids,
            select_hook = self.on_entity_id_selection,
            sort        = True,
        )
        # fmt: on

        slw = ux_make_scrollable(list_w, h=ux.SCROLL.OFF)

        left_lay.add_widget(slw)

        sp.set_style_sheet('QSplitter::handle { background-color: darkgrey; }')
        sp.set_handle_width(1)
        sp.add_widget(left_w)
        sp.add_widget(ux.Widget())
        sp.set_stretch_factor(0, 1)
        sp.set_stretch_factor(1, 2)  # -- TODO: set it for extra panes as needed

        self.main_w = sp
        return sp

    def set_pane(self, index: int, w: ux.Widget):
        main_w = self.main_w
        assert main_w, f'{self.__class__} - no widget has been created yet'

        num_panes = self.num_panes
        if num_panes and 0 <= index < num_panes:
            if index < num_panes:
                main_w.replace_widget(index + 1, w)
            else:
                main_w.add_widget(w)
                self.num_panes = num_panes + 1

    def on_entity_id_selection(self, id_value: str):
        if self.num_panes:
            obj: Traitable = self.coll.cls(_id=ID(id_value))
            se = self.stocker
            ed = se or TraitableEditor.editor(obj)
            self.current_entity = obj
            self.current_editor = ed
            w = ed.main_widget()
            self.set_pane(0, w)

    def on_new_entity(self, flag):
        cls = self.current_class
        if cls:
            new_entity = cls()
            ed = TraitableEditor.editor(new_entity)

            def accept_hook(rc: RC):
                if not rc:
                    ux_warning(rc.error(), parent=self.main_w, on_close=lambda ctx: None)
                    # -- TODO: should we "merge" values from the existing instance?

                else:
                    self.searchable_list.add_choice(new_entity.id().value)

            ed.popup(copy_entity=False, title=f'New Entity of {cls.__name__}', save=True, accept_hook=accept_hook)

    def on_deleted_entity(self, deleted_entity: Traitable):
        self.searchable_list.remove_choice(deleted_entity.id().value)
        if self.current_entity is deleted_entity:
            self.set_pane(0, ux.Widget())



from py10x_kernel import BProcessContext
from py10x_kernel import BTraitableProcessor as BTP  # noqa: N817

# TODO: consider splitting rename DEBUG into TYPE_CHECK and EAGER_LOAD.
# TODO: consider making CONVERT/TYPE_CHECK mode default.


def CHANGE_MODE(debug: bool = -1, convert_values: bool = -1):  # noqa: N802
    return BTP.create(-1, convert_values, debug, True, False)


def DEFAULT_CACHE(debug: bool = -1, convert_values: bool = -1):  # noqa: N802
    return BTP.create(0, convert_values, debug, False, True)


def GRAPH_ON(debug: bool = -1, convert_values: bool = -1):  # noqa: N802
    return BTP.create(1, convert_values, debug, False, False)


def GRAPH_OFF(debug: bool = -1, convert_values: bool = -1):  # noqa: N802
    return BTP.create(0, convert_values, debug, False, False)


def DEBUG_ON(convert_values: bool = -1):  # noqa: N802
    return BTP.create(-1, convert_values, 1, True, False)


def DEBUG_OFF(convert_values: bool = -1):  # noqa: N802
    return BTP.create(-1, convert_values, 0, True, False)


def CONVERT_VALUES_ON(debug: bool = -1):  # noqa: N802
    return BTP.create(-1, 1, debug, True, False)


def CONVERT_VALUES_OFF(debug: bool = -1):  # noqa: N802
    return BTP.create(-1, 0, debug, True, False)


def INTERACTIVE():  # noqa: N802
    return BTP.create_interactive()


# noinspection PyMethodOverriding
class ProcessContext(BProcessContext):
    @staticmethod
    def set_flags(flags: int) -> int:
        old_flags = BProcessContext.BPC.flags()
        BProcessContext.BPC.set_flags(flags)
        return old_flags

    @staticmethod
    def reset_flags(flags: int) -> int:
        old_flags = BProcessContext.BPC.flags()
        BProcessContext.BPC.reset_flags(flags)
        return old_flags

    @staticmethod
    def replace_flags(flags: int) -> int:
        old_flags = BProcessContext.BPC.flags()
        BProcessContext.BPC.replace_flags(flags)
        return old_flags

    @staticmethod
    def flags() -> int:
        return BProcessContext.BPC.flags()


class FlagsContext:
    s_default_flags = 0

    def __init__(self, flags=None):
        self._flags = flags if flags is not None else self.s_default_flags

    def __enter__(self):
        self._old_flags = ProcessContext.set_flags(self._flags)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        ProcessContext.BPC.replace_flags(self._old_flags)


class CACHE_ONLY(FlagsContext):
    s_default_flags = ProcessContext.CACHE_ONLY



import rio


class GroupBox(rio.Component):
    """
    A component that groups related controls with a labeled border.

    Attributes:
        title: The text displayed as the group's title.
        children: A list of child components to be grouped.
    """

    children: list[rio.Component] = []
    title: str = ''

    def build(self) -> rio.Component:
        """
        Build the UI as a fieldset with a legend title and child components.
        """
        return rio.Stack(
            rio.Rectangle(
                fill=rio.Color.TRANSPARENT,
                stroke_width=0.1,
                stroke_color=rio.Color.GRAY,
                corner_radius=0.5,
            ),
            rio.Column(
                rio.Text(self.title, margin_bottom=-0.5),  # Title sits on the border
                rio.Column(*self.children, margin=1),  # Content inside
            ),
        )



import pytest
from core_10x.named_constant import Enum, EnumBits, ErrorCode, NamedConstant, NamedConstantTable
from core_10x.traitable import T, Traitable


class Status(NamedConstant):
    ACTIVE = ()
    INACTIVE = ()


class StatusEx(Status):
    DEPRECATED = ()


class X(Traitable):
    status: Status = T()


@pytest.mark.parametrize('status', [Status.ACTIVE, StatusEx.DEPRECATED])
def test_serialize(ts_instance, status):
    with ts_instance:
        x = X(status=status)
        s = x.serialize_object()
        status_val = {'_type': '_nx', '_cls': 'test_named_constant/StatusEx', '_obj': 'DEPRECATED'} if status is StatusEx.DEPRECATED else 'ACTIVE'
        assert s == {'_id': x.id().value, '_rev': 0, 'status': status_val}

        x.save()

        assert x == X.load(x.id())


def test_named_constant_value_and_table():
    class Row(NamedConstant):
        A = ()
        B = ()

    class Col(NamedConstant):
        X = ()
        Y = ()

    table = NamedConstantTable(
        Row,
        Col,
        A=(1, 2),
        B=(3, 4),
    )

    assert isinstance(table, NamedConstantTable)
    # Access by constant and by name
    assert table[Row.A]['X'] == 1
    assert table['A']['Y'] == 2
    assert table[Row.B]['X'] == 3
    assert table['B']['Y'] == 4

    # primary_key lookup by secondary key/value
    assert table.primary_key('Y', 2) == Row.A
    assert table.primary_key('Y', 4) == Row.B
    assert table.primary_key('Y', 999) is None


def test_enum_and_enum_bits_operations():
    class Priority(Enum, seed=1):
        LOW = ()
        MEDIUM = ()
        HIGH = ()

    assert int(Priority.LOW.value) == 1
    assert int(Priority.MEDIUM.value) == 2
    assert int(Priority.HIGH.value) == 3

    class Permissions(EnumBits):
        READ = ()
        WRITE = ()
        EXECUTE = ()

    read = Permissions.READ
    write = Permissions.WRITE
    both = read | write
    assert isinstance(both, Permissions)
    # bitwise ops may return equivalent-but-not-identical instances; compare by value/name
    assert (both & read).value == read.value
    assert (both & write).value == write.value
    assert Permissions.names_from_value(both.value) == ('READ', 'WRITE')

    # round trips from string/int/tuple via from_any_xstr
    assert Permissions.from_any_xstr(both.value).value == both.value
    assert Permissions.from_any_xstr(('READ', 'WRITE')).value == both.value
    none_const = Permissions.NONE
    assert Permissions.from_any_xstr(()).value == none_const.value


def test_error_code_formatting():
    class MyErrors(ErrorCode):
        NOT_FOUND = 'Item {0} not found'
        INVALID_VALUE = 'Invalid value {value}'

    assert MyErrors.NOT_FOUND(42) == 'Item 42 not found'
    assert MyErrors.INVALID_VALUE(value='X') == 'Invalid value X'



"""
Test documentation examples by extracting and running code blocks at runtime.

This module automatically extracts Python code blocks from documentation files
and runs them as tests to ensure examples remain functional.
"""

from __future__ import annotations

import ast
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from core_10x.ts_store import TsStore

if TYPE_CHECKING:
    from collections.abc import Generator

# Add the project root to Python path for imports
project_root = Path(__file__).parent.parent.parent
sys.path.insert(0, str(project_root))

# Module-level variable for documentation files to avoid duplication
DOCUMENTATION_FILES = [
    'README.md',
    'GETTING_STARTED.md',
    'INSTALLATION.md',
    'CONTRIBUTING.md',
]
MISSING_DOCUMENTATION_FILES = [doc_file for doc_file in DOCUMENTATION_FILES if not (project_root / doc_file).exists()]

pytestmark = pytest.mark.skipif(
    bool(MISSING_DOCUMENTATION_FILES),
    reason=(
        'Documentation files are not available in this installation: '
        f'{", ".join(MISSING_DOCUMENTATION_FILES)}'
    ),
)


def extract_code_blocks_from_file(filepath: Path) -> Generator[tuple[str, str], None, None]:
    """Extract Python code blocks from a markdown file."""
    content = filepath.read_text(encoding='utf-8')

    # Find code blocks (```python ... ```)
    pattern = r'```python\s*\n(.*?)\n```'
    matches = re.findall(pattern, content, re.DOTALL)

    for i, code_block in enumerate(matches, 1):
        # Clean up the code block
        code_block = code_block.strip()

        # Skip empty blocks
        if not code_block:
            continue

        # Generate a test name from the file and block number
        test_name = f'{filepath.stem}_block_{i}'
        yield test_name, code_block


def extract_code_blocks_from_docs() -> Generator[tuple[str, str, str], None, None]:
    """Extract all Python code blocks from documentation files."""
    docs_dir = project_root

    for doc_file in DOCUMENTATION_FILES:
        filepath = docs_dir / doc_file
        if not filepath.exists():
            continue
        for test_name, code_block in extract_code_blocks_from_file(filepath):
            yield f'{test_name}', code_block, doc_file


def is_ui_code_block(code_block: str) -> bool:
    """Check if a code block contains UI-related imports or usage."""
    ui_indicators = [
        'import ui_10x',
        'from ui_10x',
        'ui_10x.',
    ]
    return any(indicator in code_block for indicator in ui_indicators)


def validate_python_syntax(code: str) -> bool:
    """Validate that code has valid Python syntax."""
    try:
        ast.parse(code)
        return True
    except SyntaxError:
        return False



@pytest.fixture(autouse=True)
def patch_person():
    import core_10x.code_samples.person
    old_history_class = core_10x.code_samples.person.Person.s_history_class
    core_10x.code_samples.person.Person.s_history_class = None
    yield
    core_10x.code_samples.person.Person.s_history_class = old_history_class

@pytest.mark.parametrize(
    'test_name,code_block,future_annotations',
    [
        (name, code, future_annotations)
        for name, code, src in extract_code_blocks_from_docs()
        for future_annotations in (True, False)
        if not is_ui_code_block(code)  # Skip UI code blocks - tested separately
    ],
)
def test_documentation_code_block_execution(test_name: str, code_block: str, future_annotations: bool):
    """Test that documentation code blocks can execute successfully."""
    # Skip if code block is empty
    if not code_block.strip():
        pytest.skip('Empty code block')

    # Validate syntax
    assert validate_python_syntax(code_block), f'Syntax error in {test_name}'

    doc_file_name = test_name.split('_block_')[0]
    fake_module_name = f'__doc_test_{doc_file_name}__'
    if not (fake_module:=sys.modules.get(fake_module_name)):
        fake_module = type(sys)('module')
        fake_module.__dict__.update(
            {
                '__name__': fake_module_name,
                '__file__': f'<{doc_file_name}>',
                '__builtins__': __builtins__,
            }
        )
        sys.modules[fake_module_name] = fake_module

    if future_annotations:
        exec('from __future__ import annotations', fake_module.__dict__)

    try:
        exec(code_block, fake_module.__dict__)
    finally:
        # Clean up the fake module
        if fake_module_name in sys.modules:
            del sys.modules[fake_module_name]

        TsStore.s_instances.clear()


@pytest.mark.parametrize(
    'test_name,code_block,source_file',
    [
        (f'{src}_{name}', code, src)
        for name, code, src in extract_code_blocks_from_docs()
        if is_ui_code_block(code)  # Only test UI code blocks
    ],
)
def test_documentation_ui_code_block_syntax(test_name: str, code_block: str, source_file: str):
    """Test that UI documentation code blocks have valid syntax (execution tested separately in UI tests)."""
    # Skip if code block is empty
    if not code_block.strip():
        pytest.skip('Empty code block')

    # Validate syntax only for UI blocks (execution requires async UI environment)
    assert validate_python_syntax(code_block), f'Syntax error in UI code block {source_file} {test_name}'


def test_documentation_files_and_code_blocks():
    """Test that documentation files exist and contain Python code blocks."""
    docs_dir = project_root
    total_blocks = 0
    ui_blocks = 0
    core_blocks = 0

    # Check that all expected documentation files exist
    for filename in DOCUMENTATION_FILES:
        filepath = docs_dir / filename
        assert filepath.exists(), f'Documentation file {filename} not found'
        assert filepath.stat().st_size > 0, f'Documentation file {filename} is empty'

    # Check that files contain Python code blocks
    for _name, code, _src in extract_code_blocks_from_docs():
        total_blocks += 1
        if is_ui_code_block(code):
            ui_blocks += 1
        else:
            core_blocks += 1

    # Ensure we found some code blocks
    assert total_blocks > 0, 'No Python code blocks found in documentation'
    assert core_blocks > 0, 'No core Python code blocks found in documentation'

    # We expect some UI blocks but most should be core functionality
    assert core_blocks >= ui_blocks, f'Expected more core blocks ({core_blocks}) than UI blocks ({ui_blocks})'



from core_10x.rc import RC, RC_TRUE, Enum, ErrorCode


class CONDITION(Enum):
    BEGIN   = ()
    MAIN    = ()
    END     = ()

class A_PROBLEM(ErrorCode):
    DOES_NOT_EXIST  = 'Entity {cls}.{id} does not exist'
    REV_CONFLICT    = 'Entity {cls}.{id} - revision {rev} is outdated'
    SAVE_FAILED     = 'Failed to save entity {cls}.{id}'

rc = RC_TRUE
print(rc)

try:
    rc.add_error('error')
    raise AssertionError('must have thrown')
except Exception as e:
    print(str(e))

rc = RC(CONDITION.MAIN, dict( a = 1, b = 1))
print(rc)

rc.add_data([100, 200])
print(rc)

rc1 = RC(False, 'just an error')
print(rc1)
try:
    rc1.throw()
except Exception as e:
    print(str(e))

rc = RC(A_PROBLEM.REV_CONFLICT, dict(cls = 'XXX', id = '123456', rev = 3))
print(rc)

rc = RC(True)
rc.add_error('First error')
rc.add_error(rc1)
print(rc)


from __future__ import annotations

from core_10x.named_constant import EnumBits

import ui_10x.platform_interface as i


class Style(i.Style):
    class EnumMeta(type):
        def __getattr__(cls, value):
            if value != value.upper():
                return getattr(cls, value.upper()).value

    class StandardPixmap(EnumBits, metaclass=EnumMeta):
        """Mapping of Qt QStyle::StandardPixmap values to Material Icons."""

        # Dialog Buttons
        SP_DIALOGAPPLYBUTTON = ('done',)
        SP_DIALOGCANCELBUTTON = ('close',)
        SP_DIALOGCLOSEBUTTON = ('close',)
        SP_DIALOGDISCARDBUTTON = ('close',)
        SP_DIALOGHELPBUTTON = ('help_outline',)
        SP_DIALOGNOBUTTON = ('close',)
        SP_DIALOGOKBUTTON = ('check',)
        SP_DIALOGOPENBUTTON = ('folder_open',)
        SP_DIALOGRESETBUTTON = ('restart_alt',)
        SP_DIALOGSAVEBUTTON = ('save',)
        SP_DIALOGYESBUTTON = ('check',)
        # Arrows and Navigation
        SP_ARROWBACK = ('arrow_back',)
        SP_ARROWDOWN = ('arrow_downward',)
        SP_ARROWFORWARD = ('arrow_forward',)
        SP_ARROWLEFT = ('arrow_left',)
        SP_ARROWRIGHT = ('arrow_right',)
        SP_ARROWUP = ('arrow_upward',)
        # File System and Folders
        SP_DIRCLOSEDICON = ('folder',)
        SP_DIRHOMEICON = ('home',)
        SP_DIRICON = ('folder',)
        SP_DIRLINKICON = ('folder_special',)
        SP_DIROPENICON = ('folder_open',)
        SP_FILEDIALOGBACK = ('arrow_back',)
        SP_FILEDIALOGCONTENTSVIEW = ('view_list',)
        SP_FILEDIALOGDETAILEDVIEW = ('grid_view',)
        SP_FILEDIALOGEND = ('last_page',)
        SP_FILEDIALOGINFOVIEW = ('info',)
        SP_FILEDIALOGLISTVIEW = ('list',)
        SP_FILEDIALOGNEWFOLDER = ('create_new_folder',)
        SP_FILEDIALOGSTART = ('first_page',)
        SP_FILEDIALOGTOPARENT = ('arrow_upward',)
        SP_FILEICON = ('description',)
        SP_FILELINKICON = ('insert_link',)
        # Drives and Devices
        SP_COMPUTERICON = ('computer',)
        SP_DESKTOPICON = ('desktop_windows',)
        SP_DRIVECDICON = ('album',)
        SP_DRIVEDVDICON = ('album',)
        SP_DRIVEFDICON = ('save',)
        SP_DRIVEHDICON = ('storage',)
        SP_DRIVENETICON = ('cloud',)
        SP_HOMEICON = ('home',)
        SP_TRASHICON = ('delete',)
        # Media Controls
        SP_MEDIAPAUSE = ('pause',)
        SP_MEDIAPLAY = ('play_arrow',)
        SP_MEDIASEEKBACKWARD = ('fast_rewind',)
        SP_MEDIASEEKFORWARD = ('fast_forward',)
        SP_MEDIASKIPBACKWARD = ('skip_previous',)
        SP_MEDIASKIPFORWARD = ('skip_next',)
        SP_MEDIASTOP = ('stop',)
        SP_MEDIAVOLUME = ('volume_up',)
        SP_MEDIAVOLUMEMUTED = ('volume_off',)
        # Message Boxes
        SP_MESSAGEBOXCRITICAL = ('error',)
        SP_MESSAGEBOXINFORMATION = ('info',)
        SP_MESSAGEBOXQUESTION = ('help',)
        SP_MESSAGEBOXWARNING = ('warning',)
        # Browser Controls
        SP_BROWSERRELOAD = ('refresh',)
        SP_BROWSERSTOP = ('stop',)
        # Title Bar and Window Controls
        SP_TITLEBARCLOSEBUTTON = ('close',)
        SP_TITLEBARCONTEXTHELPBUTTON = ('help_outline',)
        SP_TITLEBARMAXBUTTON = ('maximize',)
        SP_TITLEBARMENUBUTTON = ('menu',)
        SP_TITLEBARMINBUTTON = ('minimize',)
        SP_TITLEBARNORMALBUTTON = ('restore',)
        SP_TITLEBARSHADEBUTTON = ('expand_less',)
        SP_TITLEBARUNSHADEBUTTON = ('expand_more',)
        # Toolbar and Dock Widgets
        SP_DOCKWIDGETCLOSEBUTTON = ('close',)
        SP_TOOLBARHORIZONTALEXTENSIONBUTTON = ('chevron_right',)
        SP_TOOLBARVERTICALEXTENSIONBUTTON = ('chevron_down',)
        # Platform-Specific
        SP_COMMANDLINK = ('arrow_right_alt',)
        SP_VISTASHIELD = ('security',)

    def standard_icon(self, style_icon: int):
        return f'material/{self.StandardPixmap.s_reverse_dir[style_icon].label}'


class Application(i.Application):
    @classmethod
    def instance(cls) -> Application:
        raise NotImplementedError()

    @classmethod
    def style(cls):
        return Style()



from __future__ import annotations

from typing import TYPE_CHECKING

from core_10x.exec_control import INTERACTIVE
from core_10x.global_cache import cache
from core_10x.py_class import PyClass
from core_10x.rc import RC_TRUE
from core_10x.trait import Ui
from core_10x.traitable import Traitable

from ui_10x.trait_editor import TraitEditor
from ui_10x.traitable_view import TraitableView
from ui_10x.utils import UxDialog, ux, ux_warning

if TYPE_CHECKING:
    from collections.abc import Callable

    from core_10x.rc import RC


class TraitableEditor:
    """
    For many a Traitable, a default Editor is sufficient.
    A custom Editor, if needed, should be in a particular class, module and package.
    Suppose your Traitable subclass is abc.zyz.messages.TextMessage
    - the custom editor class must be TextMessageEditor
    - its module name must be messages_ui
    - the module may be in either abc.xyz, abc.xyz.ui or alternative packages
    - an editor class is searched first in alternative packages, if any; then in abc.xyz and abc.xyz.ui
    - if a custom class is not found, the default TraitableEditor is used.
    """

    @staticmethod
    @cache
    def find_editor_class(
        traitable_class: type[Traitable], *alternative_packages, traitable_class_parent: type[Traitable] = None, verify_custom_class: bool = True
    ):
        assert issubclass(traitable_class, Traitable), f'{traitable_class} is not a subclass of Traitable'
        assert not traitable_class_parent or issubclass(traitable_class, traitable_class_parent), (
            f'{traitable_class} is not a subclass of {traitable_class_parent}'
        )

        found = PyClass.find_related_class(traitable_class, 'ui', 'Editor', *alternative_packages, alternative_parent_class=traitable_class_parent)
        if found:
            if verify_custom_class:
                assert issubclass(found, TraitableEditor), (
                    f'{traitable_class} has a custom editor class {found} which is not a subclass of TraitableEditor'
                )

            return found

        return TraitableEditor

    @staticmethod
    def editor(
        entity: Traitable,
        *alternative_packages,                  #-- if custom editor class should be looked up somewhere else
        traitable_class_parent      = None,     #-- if a custom editor of entity's parent class could be used if it's missing for the entity.__class__
        verify_custom_class: bool   = True,     #-- if a custom editor class should be verified if exists
        view: TraitableView         = None,     #-- if a specific view for the entity should be used
        read_only: bool             = False     #-- browser if True
    ) -> TraitableEditor:
        editor_class = TraitableEditor.find_editor_class(
            entity.__class__,
            *alternative_packages,
            traitable_class_parent  = traitable_class_parent,
            verify_custom_class     = verify_custom_class
        )
        return editor_class(entity, view = view, read_only = read_only, _confirm = True)

    @staticmethod
    def left_mouse_callback_name(trait_name: str) -> str:   return trait_name + '_left_mouse_cb'

    @staticmethod
    def right_mouse_callback_name(trait_name: str) -> str:  return trait_name + '_right_mouse_cb'

    s_left_mouse_callbacks = {}     #-- { trait_name: <lm_cb> }  # def trait_name_left_mouse_cb(table_view, traitable): ...
    s_right_mouse_callbacks = {}    #-- { trait_name: <rm_cb> }  # def trait_name_right_mouse_cb(table_view, traitable): ...
    @classmethod
    def set_mouse_callbacks(cls, trait_name: str):
        lmc = getattr(cls, cls.left_mouse_callback_name(trait_name), None)
        if lmc:
            cls.s_left_mouse_callbacks[trait_name] = lmc
        rmc = getattr(cls, cls.right_mouse_callback_name(trait_name), None)
        if rmc:
            cls.s_right_mouse_callbacks[trait_name] = rmc

    @classmethod
    def left_mouse_callback(cls, trait_name: str):
        lmc = cls.s_left_mouse_callbacks.get(trait_name)
        if not lmc:
            lmc = cls.s_left_mouse_callbacks.get('')
        return lmc

    @classmethod
    def right_mouse_callback(cls, trait_name: str):
        rmc = cls.s_right_mouse_callbacks.get(trait_name)
        if not rmc:
            rmc = cls.s_right_mouse_callbacks.get('')
        return rmc

    s_traitable_class: type[Traitable] = None
    def __init_subclass__(cls, traitable_class: type[Traitable] = None):
        if traitable_class:
            assert issubclass(traitable_class, Traitable), 'traitable_class must be a subclass of Traitable'

            cls.s_traitable_class = traitable_class

            #-- Mouse callbacks, if any
            cls.s_left_mouse_callbacks = {}
            cls.s_right_mouse_callbacks = {}

            cls.set_mouse_callbacks('')     #-- common mouse callbacks
            for trait_name in traitable_class.s_dir:
                cls.set_mouse_callbacks(trait_name)

    def __init__(self, entity: Traitable, view: TraitableView = None, read_only = False, _confirm = False):
        assert _confirm, 'Do not call TraitableEditor() directly, use TraitableEditor.editor() instead'
        assert isinstance(entity, Traitable), 'entity must be an instance of Traitable'

        entity_class = entity.__class__
        if view is None:
            view = TraitableView.default(entity_class, read_only = read_only)
        else:
            assert issubclass(entity_class, view.cls), f'Given view is not for {entity_class}'

        self.entity = entity
        self.traitable_processor = None

        trait_dir = entity_class.s_dir
        self.trait_hints = trait_hints = {trait_dir[trait_name]: ui_hint for trait_name, ui_hint in view.ui_hints.items() if not ui_hint.flags_on(Ui.HIDDEN)}

        self.main_w: ux.Widget|None = None
        self.callbacks_for_traits = {}
        self.init()

        callbacks = self.callbacks_for_traits
        self.trait_editors = {
            trait.name: TraitEditor(entity, trait, ui_hint, custom_callback = callbacks.get(trait.name), traitable_processor = lambda: self.traitable_processor)
            for trait, ui_hint in trait_hints.items()
        }

    def init(self):
        ...

    def set_callback_for_trait(self, trait_name: str, bound_method):
        self.callbacks_for_traits[trait_name] = bound_method

    def callback_for_trait(self, trait_name: str):
        return self.callbacks_for_traits.get(trait_name)

    def main_layout(self) -> ux.Layout:
        lay = ux.FormLayout()

        te: TraitEditor
        for te in self.trait_editors.values():
            label = te.new_label()
            w = te.new_widget()
            lay.add_row(label, w)
            if te.ui_hint.flags_on(Ui.SEPARATOR):
                lay.add_row(ux.separator())

        return lay

    def row_layout(self) -> ux.Layout:
        row = ux.HBoxLayout()

        stretched_trait_found = False
        for te in self.trait_editors.values():
            stretch = te.ui_hint.param('stretch', 0)
            if stretch:
                stretched_trait_found = True

            row.add_widget(te.new_label())
            row.add_widget(te.new_widget(), stretch=stretch)

            if te.ui_hint.flags_on(Ui.SEPARATOR):
                row.add_widget(ux.separator(horizontal = False))

        if stretched_trait_found:
            row.add_widget(ux.Label(), stretch = 1)

        return row

    def main_widget(self) -> ux.Widget:
        self.main_w = w = ux.Widget()
        lay = self.main_layout()
        w.set_layout(lay)
        return w

    def _cleanup_tp(self, apply: bool):
        self.main_w = None
        if self.traitable_processor:
            if apply:
                self.traitable_processor.export_nodes()
            self.traitable_processor=None

    def _dialog(self, layout: ux.Layout, title: str, ok: str, min_width: int, on_accept: Callable[[], RC]) -> UxDialog:
        ux.init()
        if layout is not None:
            w = self.main_w = ux.Widget()
            w.set_layout(layout)
        else:
            w = self.main_widget()

        def accept_callback():
            self._cleanup_tp(True)
            rc = self.entity.verify()
            if rc:
                return on_accept()
            return rc

        def cancel_callback():
            self._cleanup_tp(False)

        return UxDialog(w, title = title, accept_callback = accept_callback, cancel_callback = cancel_callback, ok = ok, min_width = min_width)

    def dialog(self, layout: ux.Layout = None, copy_entity: bool = True, title: str = '', save: bool = False, accept_hook: Callable[[RC],None]  = None, min_width: int = 0) -> UxDialog:
        if title is None:
            title = ''
        elif not title:
            title = self.entity.__class__.__name__

        ok = 'Save' if save else 'Ok'

        def on_accept():
            rc = self.entity.save() if save else RC_TRUE
            if not rc:
                self.warning(rc.error())
            if accept_hook:
                accept_hook(rc)
            return rc

        if copy_entity:
            self.traitable_processor = INTERACTIVE()

        return self._dialog(layout, title, ok, min_width, on_accept=on_accept)

    def popup(self, layout: ux.Layout = None, copy_entity = True, title: str = '', save: bool = False, accept_hook: Callable[[RC],None] = None, min_width: int = 0) -> None:
        self.dialog(layout = layout, copy_entity = copy_entity, title = title, save = save, accept_hook = accept_hook, min_width = min_width).exec()

    def warning(self, msg: str, title: str = ''):
        ux_warning(msg, parent = self.main_w, title = title, on_close=lambda ctx: None)



import inspect
import ast
import textwrap
from typing import Any
import numba

from core_10x.trait import Trait
from core_10x.traitable import Traitable, T, RT

class NodeTransforfmer(ast.NodeTransformer):
    def __init__(self, method_name: str):
        self.method_name = method_name
        self.self_traits = []
        self.params = []

    def visit_FunctionDef(self, node):
        node.name = self.method_name

        node = self.generic_visit(node)

        node.args = ast.arguments(
            posonlyargs = [],
            args        = [ast.arg(arg = p) for p in self.params],
            vararg      = None,
            kwonlyargs  = [],
            kw_defaults = [],
            kwarg       = None,
            defaults    = []
        )
        return node

    def visit_Attribute(self, node):
        node = self.generic_visit(node)

        if isinstance(node, ast.Attribute):
            if isinstance(node.value, ast.Name) and node.value.id == 'self':
                trait_name = node.attr
                if not trait_name in self.self_traits:
                    self.self_traits.append(trait_name)
                    param = f'self_{trait_name}'
                    self.params.append(param)
                    return ast.copy_location(ast.Name(id = param, ctx = ast.Load()), node)

        return node

class GetterCompiler(Traitable):
    s_target_getter_suffix = 'numba'

    traitable_class: type[Traitable]    = RT()
    trait_name: str                     = RT()

    original_getter: Any                = RT()
    target_getter_name: str             = RT()

    ast_node_transformer: NodeTransforfmer  = RT(T.STICKY)

    target_ast_tree: Any                = RT()
    target_getter: Any                  = RT()
    compiled_target_getter: Any         = RT()
    modified_getter: Any                = RT()


    def original_getter_get(self):
        return self.traitable_class.trait(self.trait_name).f_get

    def target_getter_name_get(self) -> str:
        return f'{self.trait_name}_get_{self.__class__.s_target_getter_suffix}'

    def ast_node_transformer_get(self) -> NodeTransforfmer:
        return NodeTransforfmer(self.target_getter_name)

    def target_getter_get(self):
        src = inspect.getsource(self.original_getter)
        src = textwrap.dedent(src)
        tree = ast.parse(src)

        transformer = self.ast_node_transformer
        new_tree = transformer.visit(tree)
        ast.fix_missing_locations(new_tree)
        self.target_ast_tree = new_tree

        code = compile(new_tree, '<jit>', 'exec')

        g = self.original_getter.__globals__
        target_getter_name = self.target_getter_name
        ns = {}
        exec(code, g, ns)
        target_getter = ns[target_getter_name]
        setattr(self.traitable_class, target_getter_name, target_getter)

        return target_getter

    def compiled_target_getter_get(self):
        target_getter = numba.njit(self.target_getter)
        return target_getter

    def modified_getter_get(self):
        param_trait_names = self.ast_node_transformer.self_traits
        compiled_getter = self.compiled_target_getter
        def _mod_getter(obj):
            args = tuple(obj.get_value(name) for name in param_trait_names)
            return compiled_getter(*args)
        return _mod_getter

    def print_target_getter(self):
        print(ast.unparse(self.target_ast_tree))



from core_10x.traitable import RC, RC_TRUE, RT, T, Traitable


class Cross(Traitable):
    cross: str = RT(T.ID) // 'e.g., GBP/USD or CHF/JPY'

    base_ccy: str = RT(T.ID_LIKE) // 'Base (left) currency'
    quote_ccy: str = RT(T.ID_LIKE) // 'Quote (right) currency'

    def cross_get(self) -> str:
        return f'{self.base_ccy}/{self.quote_ccy}'

    def cross_set(self, trait, cross: str) -> RC:
        parts = cross.split('/')
        assert len(parts) == 2
        c1, c2 = parts
        rc1 = self.set_value('base_ccy', c1)
        rc2 = self.set_value('quote_ccy', c2)
        if rc1 and rc2:
            return RC_TRUE
        rc = RC(True)
        rc <<= rc1
        rc <<= rc2
        return rc


if __name__ == '__main__':
    cross = 'A/B'

    # -- Does not complain that non-ID traits can't be set during initialization.
    # -- Note the T.ID_LIKE declaration!
    c1 = Cross(base_ccy='A', quote_ccy='B')

    # -- Apparently, after setting base_ccy and quote_ccy, the cross is correct, but the id is wrong!
    # -- This is also fixed by the T.ID_LIKE declaration!
    c2 = Cross(_replace=True, base_ccy='A', quote_ccy='B')
    assert c2.id().value == 'A/B'

    # #-- Case 1 - doesn't allow to create the second instance claiming it sets non-ID traits
    # c1 = Cross(cross = cross)
    # try:
    #     c2 = Cross(cross = cross)
    # except Exception as e:
    #     print(e)
    #
    # #-- Case 2 - update is "working" exactly as the ctor
    # try:
    #     c3 = Cross.update(cross = cross)
    # except Exception as e:
    #     print(e)
    #
    # #-- Case 3: update complains about not setting at least one ID-trait - WRONG!
    # try:
    #     c4 = Cross.update(base_ccy = 'A', quote_ccy = 'B')
    # except Exception as e:
    #     print(e)
    #
    # #-- Case 4:



"""Tests for TraitableHistory functionality using pytest and real TestStore."""

from __future__ import annotations

import time
import uuid
from contextlib import nullcontext
from datetime import date, datetime

import pytest
from py10x_kernel import BTraitableProcessor
from typing_extensions import Self

from core_10x.exec_control import CACHE_ONLY, GRAPH_OFF, GRAPH_ON, INTERACTIVE
from core_10x.py_class import PyClass
from core_10x.rc import RC, RC_TRUE
from core_10x.testlib.fixtures import with_transactions
from core_10x.traitable import AsOfContext, T, Traitable, TraitableHistory
from core_10x.traitable_id import ID
from core_10x.traitable import StorableHelper, StorableHelperWithHistory
from core_10x.ts_store import TsDuplicateKeyError


@pytest.fixture(params=[True, False], ids=['frozen', 'flowing'], autouse=True)
def clock_freezer(mocker, ts_instance, request):
    class ClockFreezer(list):
        def utcnow(self):
            tm = datetime.utcnow()
            if self:
                tm, self[0] = self[0], tm
            else:
                time.sleep(0.001)

            return tm

    frozen_now = ClockFreezer()
    if request.param:
        frozen_now.append(datetime.utcnow())
        mock_dt = mocker.patch('core_10x.testlib.test_store.datetime', autospec=True)
        mock_dt.utcnow.side_effect = lambda: frozen_now[0]
        mock_dt.side_effect = lambda *args, **kw: datetime(*args, **kw)
    yield frozen_now

class NameValueTraitableBase(Traitable):
    """Test traitable class for testing."""

    name: str = T()
    value: int = T()


class NameValueTraitableCustomCollection(NameValueTraitableBase):
    s_custom_collection = True


class PersonTraitableBase(Traitable):
    """Test person class for testing."""

    name: str = T()
    age: int = T()
    email: str = T()
    dob: date = T()
    spouse: Self = T()

    def spouse_set(self, trait, spouse) -> RC:
        self.raw_set_value(trait, spouse)
        if spouse:
            spouse.raw_set_value(trait, self)
        return RC_TRUE


NameValueTraitable = type(f'PersonTraitable#{uuid.uuid1().hex}', (NameValueTraitableBase,), {'__module__': __name__})
PersonTraitable = type(f'PersonTraitable#{uuid.uuid1().hex}', (PersonTraitableBase,), {'__module__': __name__})

globals()[NameValueTraitable.__name__] = NameValueTraitable
globals()[PersonTraitable.__name__] = PersonTraitable


@pytest.fixture
def test_collection(test_store):
    """Create a test collection for testing."""
    collection_name = f'test_collection_{uuid.uuid1()}'
    yield test_store.collection(collection_name=collection_name)
    test_store.delete_collection(collection_name=collection_name)
    test_store.delete_collection(collection_name=f'{collection_name}#history')


@pytest.fixture
def test_store(ts_instance):
    store = ts_instance
    store.username = 'test_user'
    store.begin_using()
    yield store
    for cn in store.collection_names() :
        store.delete_collection(cn)
    store.end_using()


class TestTraitableHistory:
    """Test TraitableHistory functionality."""

    def test_asof_context_enter_exit(self):
        """Test AsOfContext enter and exit."""
        as_of_time = datetime(2023, 1, 1, 12, 0, 0)

        with AsOfContext(as_of_time, [NameValueTraitable]) as context:
            assert context.as_of_time == as_of_time

    def test_asof_context_manager(self):
        """Test AsOfContext as context manager."""
        as_of_time = datetime(2023, 1, 1, 12, 0, 0)

        with AsOfContext(as_of_time, [NameValueTraitable]) as context:
            assert context.as_of_time == as_of_time

    def test_traitable_history_collection(self, test_store):
        """Test TraitableHistory collection name generation."""
        # Test with default collection name
        coll = NameValueTraitable.collection()
        history_coll = NameValueTraitable.s_history_class.collection()
        assert coll.collection_name() + '#history' == history_coll.collection_name()

    def test_traitable_history_save(self, test_store, test_collection):
        serialized_data = {'_id': 'test-123', '_rev': 1, 'name': 'Test Item', 'value': 42}

        class CustomCollectionHistory(TraitableHistory, custom_collection=True):
            s_traitable_class = Traitable

        CustomCollectionHistory(
            serialized_traitable=serialized_data,
            _traitable_rev=2,
            _collection_name=test_collection.collection_name(),
        ).save().throw()

        # Verify that the history entry was saved
        assert test_collection.count() == 1

        # Get the saved document to verify its structure
        saved_docs = list(test_collection.find())
        assert len(saved_docs) == 1

        saved_doc = saved_docs[0]
        assert saved_doc['_traitable_id'] == 'test-123'
        assert saved_doc['_traitable_rev'] == 2
        assert saved_doc['_who'] == 'test_user'
        assert saved_doc['name'] == 'Test Item'
        assert saved_doc['value'] == 42
        assert isinstance(saved_doc['_at'], datetime)

    def test_save_transactional_rollback_when_history_fails(self, test_store, test_collection, with_transactions, monkeypatch):  # noqa: F811
        """When history save fails, the main document save is rolled back (revision not applied)."""

        def _save_serialized_raise(self, coll, serialized_data, old_rev):
            rev = StorableHelper._save_serialized(self, coll, serialized_data, old_rev)
            assert obj.collection().count() == 1
            if rev > old_rev:
                raise RuntimeError('history save fails')
            return rev

        monkeypatch.setattr('core_10x.traitable.StorableHelperWithHistory._save_serialized', _save_serialized_raise)

        obj = NameValueTraitable(name='x', value=1, _replace=True)
        rc = obj.save()
        assert 'history save fails' in rc.error()
        assert obj.get_revision() == 0
        assert obj.history() == []
        assert len(obj.load_many()) == (1 - with_transactions)

    def test_save_transactional_no_rollback_when_serialization_fails(self, test_store, monkeypatch, with_transactions):
        """When a nested or main save fails, the transaction rolls back (no docs committed)."""
        def _save_serialized_raise(self, coll, serialized_data, old_rev):
            if serialized_data['_id'] == '1':
                raise RuntimeError('save error')
            return StorableHelper._save_serialized(self, coll, serialized_data, old_rev)

        monkeypatch.setattr('core_10x.traitable.StorableHelperWithHistory._save_serialized', _save_serialized_raise)
        monkeypatch.setattr('core_10x.package_refactoring.PackageRefactoring.default_class_id', lambda cls, *a, **kw: PyClass.name(cls))

        class X(Traitable):
            x: int = T(T.ID)
            y: Self = T()

        with test_store:
            assert not X.load_many()
            assert not X.history()
            obj = X(x=1, y=X(x=2), _replace=True)
            assert 'Error saving traitable: save error' in obj.save(save_references=True).error()

            assert obj.y.get_revision() == 1
            assert len(X.load_many()) == 1

            X.delete_collection()

    def test_traitable_class_history_methods(self, test_store, test_collection, clock_freezer):
        """Test Traitable class history methods."""
        # Create a test traitable and save it with history
        test_item = NameValueTraitableCustomCollection(_collection_name=test_collection.collection_name())
        test_item.name = 'Test Item'
        test_item.value = 42

        # Save the traitable (this should create history)
        test_item.save()

        assert test_item._collection_name == test_collection.collection_name()

        # Test history method
        history = NameValueTraitableCustomCollection.history(_collection_name=test_item._collection_name)
        assert len(history) == 1
        assert history[0]['_traitable_id'] == test_item.id().value
        assert history[0]['_traitable_rev'] == test_item._rev
        assert history[0]['_who'] == 'test_user'
        assert '_at' in history[0]  # Should have timestamp
        assert history[0]['name'] == 'Test Item'
        assert history[0]['value'] == 42

        ts = clock_freezer.utcnow()
        test_item.value = 43
        test_item.save()
        assert test_item._rev == 2
        assert test_item.value == 43

        assert NameValueTraitableCustomCollection.restore(test_item.id(), ts)
        assert test_item._rev == 1
        assert test_item.value == 42

    def test_history_entry_creation(self, test_store):
        """Test that history entries are created when saving traitables."""
        # Create a person
        person = PersonTraitable()
        person.name = 'John Doe'
        person.age = 30
        person.email = 'john@example.com'

        # Save the person
        person.save()

        # Check that history was created
        history_collection = test_store.collection(f'{__name__.replace(".", "/")}/{PersonTraitable.__name__}#history')
        assert history_collection.count() == 1

        # Verify history entry structure
        history_docs = list(history_collection.find())
        history_doc = history_docs[0]
        assert history_doc['_traitable_id'] == person.id().value
        # History entry is created before revision increment, so it should be 0
        assert history_doc['_traitable_rev'] == person._rev
        assert history_doc['_who'] == 'test_user'
        assert '_at' in history_doc
        assert history_doc['name'] == 'John Doe'
        assert history_doc['age'] == 30
        assert history_doc['email'] == 'john@example.com'

    def test_multiple_history_entries(self, test_store):
        """Test that multiple history entries are created for updates."""

        # Create a person
        person = PersonTraitable()
        person.name = 'John Doe'
        person.age = 30
        person.email = 'john@example.com'
        person.save()

        # Update the person
        person.age = 31
        person.save()

        # Check that two history entries were created
        history_collection = test_store.collection(f'{__name__.replace(".", "/")}/{PersonTraitable.__name__}#history')
        assert history_collection.count() == 2

        # Verify both history entries
        history_docs = list(history_collection.find())
        assert len(history_docs) == 2

        # Both should have the same traitable_id but different revs
        traitable_ids = [doc['_traitable_id'] for doc in history_docs]
        assert all(tid == person.id().value for tid in traitable_ids)

        revs = {doc['_traitable_rev'] for doc in history_docs}
        assert {1, 2} == revs

    def test_history_method(self, test_store):
        """Test the history method."""

        # Create a person
        person = PersonTraitable()
        person.name = 'John Doe'
        person.age = 30
        person.email = 'john@example.com'
        person.save()

        # Update the person
        person.age = 31
        person.save()

        # Get history
        history = PersonTraitable.history()

        # Should have 2 history entries
        assert len(history) == 2

        # Both entries should have the correct traitable_id
        for entry in history:
            assert entry['_traitable_id'] == person.id().value
            assert entry['_who'] == 'test_user'
            assert '_at' in entry

    def test_latest_revision(self, test_store):
        """Test the latest_revision method."""

        # Create a person
        person = PersonTraitable()
        person.name = 'John Doe'
        person.age = 30
        person.email = 'john@example.com'
        person.save().throw()

        # Update the person
        person.age = 31
        person.save().throw()

        assert person._rev == 2

        # Get latest revision
        latest = PersonTraitable.latest_revision(person.id())

        # Should be the latest revision
        assert latest['_traitable_id'] == person.id().value
        assert latest['_traitable_rev'] == person._rev
        assert latest['_who'] == 'test_user'
        assert '_at' in latest
        assert latest['age'] == 31  # Updated age

    def test_asof_context_manager_ops(self, test_store, clock_freezer):
        """Test AsOfContext with real data."""

        # Create a person
        person = PersonTraitable()
        person.name = 'John Doe'
        person.age = 30
        person.email = 'john@example.com'
        person.save()

        # Record the time after creation
        query_time = clock_freezer.utcnow()

        # Update the person
        person.age = 31
        person.save()

        # Load the person as of the query time (should be the original version)
        with AsOfContext(query_time):
            historical_person = PersonTraitable.load(person.id())
            assert historical_person.age == 30  # Original age
            assert historical_person.name == 'John Doe'

    def test_as_of_error_cases(self, test_store, clock_freezer):
        with BTraitableProcessor.create_root():
            # Create and save a person
            person = PersonTraitable(first_name='Alyssa', last_name='Lees', dob=date(1985, 7, 5), _replace=True)
            person.spouse = PersonTraitable(first_name='James', last_name='Bond', dob=date(1985, 7, 5), _replace=True)
            person.spouse.save().throw()
            person.save().throw()

            ts = clock_freezer.utcnow()

            # Update and save again
            person.dob = date(1985, 7, 6)
            person.spouse.dob = date(1985, 7, 6)
            person.spouse.save().throw()
            person.save().throw()

            person_id = person.id()
            print(person.serialize_object())

            person.spouse.id()

        assert PersonTraitable.existing_instance_by_id(person_id, _throw=False)

        with CACHE_ONLY():
            assert not PersonTraitable.existing_instance_by_id(person_id, _throw=False)

        for ctx in (
            INTERACTIVE,
            GRAPH_ON,
            GRAPH_OFF,
        ):
            for as_of in (
                ts,
                None,
            ):
                with ctx():
                    person1 = PersonTraitable(person_id)
                    assert person1.dob == date(1985, 7, 6)

                    person_as_of = PersonTraitable.as_of(person_id, as_of_time=ts)

                    with AsOfContext(traitable_classes=[PersonTraitable], as_of_time=as_of):
                        with pytest.raises(RuntimeError, match=r'object not usable - origin cache is not reachable'):
                            _ = person1.dob

                        with pytest.raises(RuntimeError, match=r'object not usable - origin cache is not reachable'):
                            _ = person_as_of.dob

                        person2 = PersonTraitable(person_id)
                        assert person2.dob == (date(1985, 7, 5 + (as_of is None)))
                        assert person2.spouse.dob == (date(1985, 7, 5 + (as_of is None)))

                    assert person1.spouse.dob == date(1985, 7, 6)
                    assert person_as_of.dob == date(1985, 7, 5), person_as_of.dob

                    assert person_as_of.spouse.dob == date(1985, 7, 6)  # note - since no AsOfContext was used, nested objects are not loaded "as_of".

                    with pytest.raises(RuntimeError, match=r'object not usable - origin cache is not reachable'):
                        _ = person2.dob

    def test_load_as_of(self, test_store, clock_freezer):
        """Test loading a traitable as of a specific time."""

        # Create a person
        person = PersonTraitable()
        person.name = 'John Doe'
        person.age = 30
        person.email = 'john@example.com'
        person.save()

        # Record the time after creation
        initial_time = clock_freezer.utcnow()

        # Update the person
        person.age = 31
        person.save()

        # Load the person as of the initial time
        historical_person = PersonTraitable.as_of(person.id(), initial_time)
        assert historical_person.age == 30  # Original age
        assert historical_person.name == 'John Doe'
        assert historical_person == person  # person's trait values also get updated due to shared cache

    def test_load_many_with_as_of(self, test_store, clock_freezer):
        """Test loading multiple traitables as of a specific time."""

        # Create two people
        person1 = PersonTraitable()
        person1.name = 'John Doe'
        person1.age = 30
        person1.email = 'john@example.com'
        person1.save()

        person2 = PersonTraitable()
        person2.name = 'Jane Smith'
        person2.age = 25
        person2.email = 'jane@example.com'
        person2.save()

        # Record the time after creation
        query_time = clock_freezer.utcnow()

        # Update both people
        person1.age = 31
        person1.save()

        person2.age = 26
        person2.save()

        with AsOfContext(query_time, [PersonTraitable]):
            historical_people = PersonTraitable.load_many()

            # Should have 2 people with original ages
            assert len(historical_people) == 2

            # Find the specific people
            john = next(p for p in historical_people if p.name == 'John Doe')
            jane = next(p for p in historical_people if p.name == 'Jane Smith')

            assert john.age == 30  # Original age
            assert jane.age == 25  # Original age

        assert person1.age == 31  # Current age
        assert person2.age == 26  # Current age

    def test_history_without_keep_history(self, test_store, monkeypatch):
        """Test that history is not created when keep_history is False."""
        monkeypatch.setattr('core_10x.package_refactoring.PackageRefactoring.default_class_id', lambda cls, *args, **kwargs: PyClass.name(cls))

        # Create a class without history tracking
        class NoHistoryTraitable(Traitable, keep_history=False):
            name: str = T()

        # Create and save an instance
        item = NoHistoryTraitable()
        item.name = 'Test Item'
        item.save()

        # Check that no history collection was created
        history_collection = test_store.collection(f'{__name__.replace(".", "/")}/{NoHistoryTraitable.__name__}#history')
        assert history_collection.count() == 0

    def test_asof_with_keep_history_false_is_noop(self, test_store, monkeypatch):
        """AsOfContext with a class that has keep_history=False is a no-op: current data is used."""
        monkeypatch.setattr('core_10x.package_refactoring.PackageRefactoring.default_class_id', lambda cls, *args, **kwargs: PyClass.name(cls))

        class NoHistoryTraitable(Traitable, keep_history=False):
            key: str = T(T.ID)
            value: str = T()

        item = NoHistoryTraitable(key='k1', value='v1', _replace=True)
        item.save().throw()
        original_helper = NoHistoryTraitable.s_storage_helper
        as_of_time = datetime(2020, 1, 1, 12, 0, 0)

        with pytest.raises(
            ValueError,
            match=r"<class 'core_10x.testlib.traitable_history_tests.TestTraitableHistory.test_asof_with_keep_history_false_is_noop.<locals>.NoHistoryTraitable'> is not storable or does not keep history",
        ):
            with AsOfContext(as_of_time, [NoHistoryTraitable]):
                pass
            assert type(NoHistoryTraitable.s_storage_helper) is type(original_helper)
        assert NoHistoryTraitable.existing_instance(key='k1') == item

    def test_asof_default_applies_to_all_traitable_subclasses(self, test_store, clock_freezer):
        """AsOfContext with default traitable_classes (None) applies to all Traitable subclasses via __mro__."""
        # Create and save a person
        person = PersonTraitable()
        person.name = 'Default AsOf Test'
        person.age = 25
        person.save()

        query_time = clock_freezer.utcnow()

        person.age = 26
        person.save()

        # Default: no traitable_classes => [Traitable]; PersonTraitable is a subclass so gets AsOf
        with AsOfContext(as_of_time=query_time):
            historical = PersonTraitable.load(person.id())
            assert historical is not None
            assert historical.age == 25
            assert historical.name == 'Default AsOf Test'

        # Outside context: current data
        current = PersonTraitable.load(person.id())
        assert current.age == 26

    def test_latest_revision_with_timestamp(self, test_store, clock_freezer):
        """Test latest_revision with specific timestamp."""
        # Create a person
        person = PersonTraitable()
        person.name = 'Time Test Person'
        person.age = 30
        person.save()

        # Record time after first save
        query_time = clock_freezer.utcnow()

        # Update the person
        person.age = 31
        person.save()

        # Get latest revision as of the query time
        latest = PersonTraitable.latest_revision(person.id(), timestamp=query_time)

        # Should be the first revision (age 30)
        assert latest['age'] == 30
        assert latest['_traitable_rev'] == 1

    def test_history_with_empty_collection(self, test_store):
        """Test history method with no history entries."""
        # Create a person but don't save it
        person = PersonTraitable()
        person.name = 'Unsaved Person'

        # Get history - should be empty
        history = PersonTraitable.history()
        assert len(history) == 0

    def test_latest_revision_with_nonexistent_id(self, test_store):
        """Test latest_revision with non-existent ID."""

        # Try to get latest revision for non-existent ID
        latest = PersonTraitable.latest_revision(ID('nonexistent'))
        assert latest is None

    def test_history_default_at_most_parameter(self, test_store):
        """Test that history method uses default _at_most=0 (no limit)."""
        # Create a person
        person = PersonTraitable()
        person.name = 'Test Person'
        person.age = 30
        person.save()

        # Update multiple times
        person.age = 31
        person.save()
        person.age = 32
        person.save()

        # Get all history entries (should get all 3)
        history = PersonTraitable.history()
        assert len(history) == 3

        # Test with explicit _at_most=2
        history_limited = PersonTraitable.history(_at_most=2)
        assert len(history_limited) == 2

    def test_latest_revision_with_timestamp_parameter(self, test_store, clock_freezer):
        """Test that latest_revision accepts timestamp parameter."""
        # Create a person
        person = PersonTraitable()
        person.name = 'Timestamp Test'
        person.age = 25
        person.save()

        # Record time after first save
        query_time = clock_freezer.utcnow()

        # Update the person
        person.age = 26
        person.save()

        # Test latest_revision with timestamp
        latest = PersonTraitable.latest_revision(person.id(), timestamp=query_time)
        assert latest is not None
        assert latest['age'] == 25  # Should be the first revision

        # Test latest_revision without timestamp (should get latest)
        latest_no_timestamp = PersonTraitable.latest_revision(person.id())
        assert latest_no_timestamp is not None
        assert latest_no_timestamp['age'] == 26  # Should be the latest revision

    def test_restore(self, test_store, clock_freezer):
        """Test restore method with save parameter."""
        # Create a person
        person = PersonTraitable()
        person.name = 'Restore Test'
        person.age = 30
        person.save()

        # Record time after first save
        query_time = clock_freezer.utcnow()

        # Update the person
        person.age = 31
        person.save()

        # Test restore without save (should not persist)
        result = PersonTraitable.restore(person.id(), timestamp=query_time, save=False)
        assert result is True
        assert person.age == 30

        # The person should still have the latest age
        person.reload()
        assert person.age == 31

        # Test restore with save (should persist)
        result = PersonTraitable.restore(person.id(), timestamp=query_time, save=True)
        assert result is True
        assert person.age == 30
        person.reload()
        assert person.age == 30

    def test_restore_with_nonexistent_timestamp(self, test_store):
        """Test restore method with non-existent timestamp."""
        # Create a person
        person = PersonTraitable()
        person.name = 'Restore Test'
        person.age = 30
        person.save()

        # Try to restore to a time before the person existed
        past_time = datetime(2020, 1, 1)
        result = PersonTraitable.restore(person.id(), timestamp=past_time, save=False)
        assert result is False

    def test_restore_with_nonexistent_id(self, test_store):
        """Test restore method with non-existent ID."""

        # Try to restore non-existent ID
        result = PersonTraitable.restore(ID('nonexistent'), save=False)
        assert result is False

    def test_traitable_history_deserialize(self, test_store):
        """Test TraitableHistory.deserialize method."""
        # Create a person
        person = PersonTraitable()
        person.name = 'Deserialize Test'
        person.age = 25
        person.save()

        # Get history entry
        history = PersonTraitable.history(_deserialize=True)
        assert len(history) == 1

        history_entry = history[0]

        # Test deserialize
        restored_person = history_entry.traitable

        assert restored_person.name == 'Deserialize Test'
        assert restored_person.age == 25
        assert restored_person.id().value == person.id().value

    def test_traitable_history_prepare_to_deserialize(self, test_store):
        """Test TraitableHistory.prepare_to_deserialize method."""
        # Create a person
        person = PersonTraitable()
        person.name = 'Prepare Test'
        person.age = 28
        person.save()

        # Get history entry
        history = PersonTraitable.history(_deserialize=True)
        history_entry = history[0]

        # Test prepare_to_deserialize
        prepared_data = history_entry.serialized_traitable

        # Should have the traitable data without history-specific fields
        assert prepared_data['name'] == 'Prepare Test'
        assert prepared_data['age'] == 28
        assert prepared_data['_id'] == person.id().value
        assert prepared_data['_rev'] == person._rev
        assert '_traitable_id' not in prepared_data
        assert '_traitable_rev' not in prepared_data
        assert '_who' not in prepared_data
        assert '_at' not in prepared_data

    def test_save_new_with_overwrite_parameter(self, test_store, test_collection):
        """Test TestCollection.save_new with overwrite parameter."""
        # Test save_new with MongoDB-style $set operation
        result = test_collection.save_new({'$set': {'_id': 'new-id', 'name': 'New Person', 'age': 25}})
        assert result == 1  # Should succeed for new document

        # Fails because the document already exists
        with pytest.raises(TsDuplicateKeyError):
            test_collection.save_new({'$set': {'_id': 'new-id'}})

        test_collection.save_new({'$set': {'_id': 'new-id'}}, overwrite=True)

    def test_find_without_filter_parameter(self, test_store, test_collection):
        """Test TestCollection.find without _filter parameter."""

        # Add documents using the interface method
        test_collection.save_new({'_id': 'doc1', 'name': 'Person 1', 'age': 25})
        test_collection.save_new({'_id': 'doc2', 'name': 'Person 2', 'age': 30})

        # Test find without _filter parameter
        results = list(test_collection.find())
        assert len(results) == 2  # Should find both persons

    def test_immutability_of_classes_without_history(self, test_store, test_collection):
        """Test immutability of storable classes that do not keep history."""

        class NoHistoryImmutableTraitable(NameValueTraitableCustomCollection, keep_history=False): ...

        # Classes without history should be marked immutable
        assert NoHistoryImmutableTraitable.s_history_class is None
        assert NoHistoryImmutableTraitable.s_immutable is True

        # First save should succeed and create a single record
        item = NoHistoryImmutableTraitable(_collection_name=test_collection.collection_name())
        item.name = 'First'
        rc1 = item.save()
        assert rc1
        assert item.name == 'First'
        assert item._rev == 1
        assert test_collection.count() == 1

        # Second save should fail (cannot update immutable records)
        item.name = 'Second'
        rc2 = item.save()
        assert not rc2
        assert item.name == 'Second'
        assert item._rev == 1

        # After reload, state should still reflect the original save
        item.reload()
        assert item.name == 'First'
        assert item._rev == 1

    def test_history_entries_are_immutable(self, test_store):
        """Test that individual history entries cannot be modified once written."""

        # Create and save a person twice to generate multiple history entries
        person = PersonTraitable()
        person.name = 'Immutable History'
        person.age = 30
        person.email = 'immutable@example.com'
        person.save().throw()

        person.age = 31
        person.save().throw()

        history_collection = test_store.collection(f'{__name__.replace(".", "/")}/{PersonTraitable.__name__}#history')
        assert history_collection.count() == 2

        # Load a history entry as TraitableHistory instance
        history_entries = PersonTraitable.history(_deserialize=True)
        assert len(history_entries) == 2

        entry = history_entries[0]
        original_who = entry._who

        # Attempting to modify and re-save the history entry should fail
        entry._who = 'someone-else'
        rc = entry.save()
        assert not rc

        # Reload from the store and verify that the stored entry was not changed
        reloaded_docs = list(history_collection.find())
        assert len(reloaded_docs) == 2
        assert any(doc['_who'] == original_who for doc in reloaded_docs)
        assert all(doc['_who'] != 'someone-else' for doc in reloaded_docs)

    def test_asof_find_returns_one_record_per_id(self, test_store, clock_freezer):
        """Test that the AsOf _find implementation returns at most one record per _id."""

        # Create two people and multiple history entries for each
        person1 = PersonTraitable()
        person1.name = 'Person One'
        person1.age = 30
        person1.email = 'one@example.com'
        person1.save()

        person1.age = 31
        person1.save()

        person2 = PersonTraitable()
        person2.name = 'Person Two'
        person2.age = 25
        person2.email = 'two@example.com'
        person2.save()

        person2.age = 26
        person2.save()

        # Sanity check: we indeed have multiple history records per person
        history = PersonTraitable.history()
        assert len(history) >= 4

        as_of_time = clock_freezer.utcnow()

        # AsOfContext uses StorableHelperAsOf._find under the hood for load_many
        with AsOfContext(as_of_time, [PersonTraitable]):
            people_as_of = PersonTraitable.load_many()

        # We should get exactly one record per traitable id
        assert len(people_as_of) == 2
        ids = [p.id().value for p in people_as_of]
        assert len(set(ids)) == len(ids)



from core_10x.directory import Directory
from core_10x.environment_variables import EnvVars
from core_10x.exec_control import INTERACTIVE
from core_10x.py_class import PyClass
from core_10x.traitable import RC, RC_TRUE, RT, AnonymousTraitable, Traitable, XNone

from ui_10x.collection_editor import Collection, CollectionEditor
from ui_10x.traitable_editor import TraitableEditor
from ui_10x.traitable_view import Ui
from ui_10x.utils import UxDialog, ux


class CEApp(Traitable):
    s_exclude_packages = ('manual_tests', 'unit_tests')

    # fmt: off
    main_store_uri: str     = RT()
    top_package: str        = RT()
    exclude_packages: list  = RT(Ui(flags = Ui.HIDDEN))
    current_class: type     = RT(Ui.choice(stretch = 1))

    my_editor: TraitableEditor  = RT()
    col_editor: CollectionEditor = RT()
    main_w: ux.Splitter         = RT()
    # fmt: on

    def main_store_uri_get(self) -> str:
        return EnvVars.main_ts_store_uri

    def main_store_uri_set(self, trait, value) -> RC:
        self.raw_set_value(trait, value)
        EnvVars.main_ts_store_uri = value
        return RC_TRUE

    def exclude_packages_get(self) -> list:
        root_name = self.top_package
        return [f'{root_name}.{name}' for name in self.s_exclude_packages]

    def current_class_set(self, trait, value) -> RC:
        self.raw_set_value(trait, value)
        coll = Collection(cls=value)
        ce = CollectionEditor(coll=coll)
        self.col_editor = ce
        w = ce.main_widget()
        self.main_w.replace_widget(0, w)

        return RC_TRUE

    def current_class_choices(self, trait) -> Directory:
        top_package = self.top_package
        if not top_package:
            return XNone

        all_traitables = PyClass.all_classes(self.top_package, exclude_packages=self.exclude_packages, parent_classes=(Traitable,))
        all_storables = tuple(cls for cls in all_traitables if cls.is_storable() and not issubclass(cls, AnonymousTraitable))
        dir = Directory(name='Classes')
        for cls in all_storables:
            path = cls.__module__.split('.')
            dir.insert(PyClass.name(cls), *path)

        return dir

    def my_editor_get(self) -> TraitableEditor:
        return TraitableEditor(self, _confirm=True)

    def main_widget(self) -> ux.Widget:
        w = ux.Widget()
        lay = ux.VBoxLayout()
        w.set_layout(lay)

        lay.add_layout(self.my_editor.row_layout())
        lay.add_widget(ux.separator())

        sp = ux.Splitter(ux.Horizontal)
        lay.add_widget(sp)

        sp.set_style_sheet('QSplitter::handle { background-color: darkgrey; }')
        sp.set_handle_width(1)

        sp.add_widget(ux.Widget())
        sp.set_stretch_factor(0, 1)
        sp.set_stretch_factor(1, 2)  # -- TODO: set it for extra panes as needed

        self.main_w = sp

        return w


if __name__ == '__main__':
    from ui_10x.apps.collection_editor_app import CEApp

    ux.init()

    top_package = input('Top level package:')
    ce_app = CEApp(top_package=top_package)

    with INTERACTIVE():
        w = ce_app.main_widget()
        d = UxDialog(w)
        d.exec()



from __future__ import annotations

import sys
import traceback
from typing import TYPE_CHECKING

from core_10x.named_constant import Enum, ErrorCode

if TYPE_CHECKING:
    from types import TracebackType


class RC:
    """
    Stands for 'Return Code'. Main usage is for functions returning 'success' or 'error' with an optional payload.

    For just a 'success':
        RC_TRUE (a constant RC(True) instance)
        return RC_TRUE

    For 'success' with a single data:
        return RC(True, data)
        RC(enum, data)      - an instance of Enum subclass with a positive value

    For 'success' with multiple data:
            rc = RC(True)
            ...
                rc.add_data(data)
            ...
            return rc

    For a single error:
        return RC(False)    - if you are in except: to catch an exception with stack info
        return RC(False, message)
        return RC(error, message)   - an instance of ErrorCode subclass (with a non-positive value)

    For multiple errors:
            rc = RC(True)
            ...
                rc.add_error(message)
            ...
            return rc
    """

    __slots__ = ('payload', 'rc')

    @classmethod
    def show_exception_info(cls, ex_info: tuple[type[BaseException], BaseException, TracebackType] = None) -> str:
        if not ex_info:
            ex_info = sys.exc_info()

        assert isinstance(ex_info, tuple) and len(ex_info) == 3, f'Invalid ex_info: {ex_info}'

        ss = traceback.StackSummary.extract(traceback.walk_tb(ex_info[2]))
        return f'{ex_info[0]} ({ex_info[1]})\n{"".join(ss.format())}'

    def __init__(self, rc, data=None):
        self.rc = rc
        self.payload = [data] if data is not None else []

    def __bool__(self):
        rc = self.rc
        dt = type(rc)
        if dt is bool:
            return rc

        if issubclass(dt, Enum):
            rc = rc.value
        else:
            assert dt is int, f'Invalid rc: {rc}'

        return rc > 0

    def __repr__(self):
        if self.__bool__():
            data = self.payload
        else:
            data = self.error()
        return f'{self.rc}: {data}'

    def __iadd__(self, err):
        return self.add_error(err)

    def __ilshift__(self, err):
        return self.add_error(err)

    def unwrap(self) -> tuple:  # -- ( rc, payload)
        return (self.rc, self.payload)

    def data(self):
        payload = self.payload
        return payload if len(payload) > 1 else payload[0]

    def error(self) -> str:
        if self.__bool__():
            return ''

        payload = self.payload
        n = len(payload)
        if n == 0:
            return self.__class__.show_exception_info()

        if n == 1:
            data = payload[0]
            if isinstance(self.rc, ErrorCode):
                return self.rc(**data)
            return data

        # -- multiple errors
        return '\n'.join(payload)

    def add_error(self, err) -> RC:
        if err is RC_TRUE:
            return self

        dt = type(err)
        if dt is str:
            self.payload.append(err)

        elif dt is RC:
            if err:
                return self
            self.payload.extend(err.payload)

        else:
            raise ValueError(f'Expected str or RC; got {type(err)}')

        if self.__bool__():
            self.rc = False

        return self

    def prepend_error_header(self, err: str) -> RC:
        """
        Inserts the first error (header), if self is already an error
        """
        if not self.__bool__():
            self.payload.insert(0, err)

        return self

    def add_data(self, data) -> RC:
        assert self.__bool__(), 'May not add_data() to an "error" RC'
        self.payload.append(data)
        return self

    def throw(self, exc: type[Exception] = RuntimeError):
        err = self.error()
        if err:
            raise exc(err)


class _RcTrue(RC):
    def __init__(self):
        super().__init__(True)

    def new_rc(self) -> RC:
        return RC(True)

    def add_error(self, err: str = ''):
        raise ValueError('May not add error to a constant RC_TRUE')

    def add_data(self, data):
        raise ValueError('May not add error to a constant RC_TRUE')


RC_TRUE = _RcTrue()



import pytest

from core_10x.environment_variables import EnvVars
from core_10x.xdate_time import XDateTime, date


def test_env_vars_date_format_applies_to_xdatetime(monkeypatch):
    # Change via environment variable and ensure it is applied on first access
    custom_fmt = '%d/%m/%Y'
    monkeypatch.setenv('XX_DATE_FORMAT', custom_fmt)

    # Accessing EnvVars.date_format should convert and apply to XDateTime
    fmt = EnvVars.date_format
    assert fmt == custom_fmt

    d = date(2024, 1, 2)
    assert XDateTime.date_to_str(d) == '02/01/2024'


@pytest.mark.parametrize(['truth','values'],[[True,['1','2','-1','True',"'quoted'"]],[False,['0','False',"''"]],[None,['','Random']]])
def test_env_vars_converts_bool_true(monkeypatch,truth,values):
    for value in values:
        for v in (value,str(value).lower()):
            monkeypatch.setenv('XX_USE_TS_STORE_TRANSACTIONS', v)
            object.__getattribute__(EnvVars, 'use_ts_store_transactions').fget.clear()
            if truth is None:
                with pytest.raises(TypeError):
                    _ = EnvVars.use_ts_store_transactions
            else:
                assert EnvVars.use_ts_store_transactions is truth, f"Expected {v} to convert to {truth}"



import pytest
import polars as pl

from core_10x.rel_db import RelDb


def test_rel_db_spec_from_uri_parses_uri_components():
    spec = RelDb.spec_from_uri("pgtest://user:pass@localhost:5432/test_db")

    assert spec.resource_class is RelDb
    assert spec.kwargs[RelDb.PROTOCOL_TAG] == "pgtest"
    assert spec.kwargs[RelDb.HOSTNAME_TAG] == "localhost"
    assert spec.kwargs[RelDb.DBNAME_TAG] == "test_db"
    assert spec.kwargs[RelDb.USERNAME_TAG] == "user"
    assert spec.kwargs[RelDb.PASSWORD_TAG] == "pass"
    assert spec.kwargs[RelDb.PORT_TAG] == 5432


def test_rel_db_query_and_insert(tmp_path):
    source_df = pl.DataFrame({'a': [1, 2, 3], 'b': ['x', 'y', 'z']})

    # Use keyword-arg construction so ResourceSpec.uri() emits the four-slash form
    # (duckdb:////abs/path) that Polars ADBC requires for absolute file paths.
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))

    with db:
        db.insert("my_table", source_df)

    # Verify in a fresh connection that the ADBC write was persisted.
    with db:
        result_df = db.query("my_table").to_polars()

    assert result_df.equals(source_df)


def test_rel_db_instance_rejects_positional_args():
    with pytest.raises(ValueError, match="keyword arguments only"):
        RelDb.instance('duckdb')


def test_rel_db_query_returns_none_for_missing_table(tmp_path):
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    with db:
        result = db.query("nonexistent_table")
    assert result is None


def test_rel_db_insert_replace(tmp_path):
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    original_df = pl.DataFrame({'a': [1, 2, 3], 'b': ['x', 'y', 'z']})
    replacement_df = pl.DataFrame({'a': [10, 20], 'b': ['p', 'q']})

    with db:
        db.insert("my_table", original_df)
    # DuckDB ADBC does not support if_exists='replace'; drop via ibis then re-insert.
    with db:
        db.drop_table("my_table")
    with db:
        db.insert("my_table", replacement_df)
    with db:
        result_df = db.query("my_table").to_polars()

    assert result_df.equals(replacement_df)


def test_rel_db_insert_append(tmp_path):
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    first_df = pl.DataFrame({'a': [1, 2], 'b': ['x', 'y']})
    second_df = pl.DataFrame({'a': [3, 4], 'b': ['z', 'w']})

    with db:
        db.insert("my_table", first_df)
    with db:
        db.insert("my_table", second_df, if_exists='append')
    with db:
        result_df = db.query("my_table").to_polars()

    expected_df = pl.concat([first_df, second_df])
    assert result_df.equals(expected_df)


def test_rel_db_drop_table(tmp_path):
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    df = pl.DataFrame({'a': [1, 2, 3]})

    # Insert via ADBC in one block, drop via ibis in the next; DuckDB requires
    # the ADBC connection to be closed before ibis can observe the committed table.
    with db:
        db.insert("my_table", df)
    with db:
        db.drop_table("my_table")
    with db:
        result = db.query("my_table")

    assert result is None


def test_rel_db_query_raises_without_context_manager(tmp_path):
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    with pytest.raises(AssertionError, match="context manager"):
        db.query("my_table")


def test_rel_db_insert_raises_without_context_manager(tmp_path):
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    df = pl.DataFrame({'a': [1]})
    with pytest.raises(AssertionError, match="context manager"):
        db.insert("my_table", df)


def test_rel_db_drop_table_raises_without_context_manager(tmp_path):
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    with pytest.raises(AssertionError, match="context manager"):
        db.drop_table("my_table")


def test_rel_db_instance_from_uri(tmp_path):
    source_df = pl.DataFrame({'val': [42]})

    # Seed the DB via instance() — it generates the four-slash absolute-path URI
    # that the ADBC driver requires.  Then verify that instance_from_uri() can
    # round-trip the same URI and reconnect for ibis-based reads.
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    with db:
        db.insert("t", source_df)

    db2 = RelDb.instance_from_uri(db._uri)
    with db2:
        result_df = db2.query("t").to_polars()

    assert result_df.equals(source_df)


def test_rel_db_connection_closed_after_context_exit(tmp_path):
    db = RelDb.instance(protocol='duckdb', dbname=str(tmp_path / 'test.db'))
    with db:
        pass
    # After exit, _connection must be cleaned up.
    assert db._connection is None



from __future__ import annotations

from ui_10x.collection_editor import Collection, CollectionEditor
from ui_10x.rio.component_builder import UserSessionContext
from ui_10x.utils import UxDialog

import rio


class CollectionEditorComponent(rio.Component):
    collection_class: type = None

    def build(self) -> rio.Component:
        with self.session[UserSessionContext]:
            return UxDialog(CollectionEditor(coll=Collection(cls=self.collection_class)).main_widget()).build(self.session)



from __future__ import annotations

import asyncio
import inspect
import operator
import types
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial, reduce
from typing import TYPE_CHECKING

from core_10x.exec_control import BTP, INTERACTIVE
from core_10x.named_constant import Enum, NamedConstant
from core_10x.traitable import Traitable

import rio
import ui_10x.platform_interface as i
from ui_10x.rio.style_sheet import StyleSheet

if TYPE_CHECKING:
    from collections.abc import Callable

    from core_10x.rc import RC
    from core_10x.ts_store import TsStore
    from py10x_kernel import BTraitableProcessor


@dataclass
class UserSessionContext:
    # TODO: backbone
    host: str
    dbname: str
    traitable_store: TsStore = None
    interactive: BTraitableProcessor = None
    authenticated: bool = False

    def begin_using(self):
        if self.traitable_store:
            self.traitable_store.begin_using()

        if not self.interactive:
            self.interactive = INTERACTIVE()
            print('interactive in', BTP.current(), self.interactive)

        self.interactive.begin_using()

    def end_using(self):
        self.interactive.end_using()
        if self.traitable_store:
            self.traitable_store.end_using()

    def __enter__(self):
        return self.begin_using()

    def __exit__(self, *args):
        self.end_using()


CURRENT_SESSION: rio.Session | None = None


@contextmanager
def session_context(session: rio.Session):
    global CURRENT_SESSION
    assert CURRENT_SESSION is None, 'Must exit from session context first! Are you using async calls in session context?'
    CURRENT_SESSION = session
    try:
        user_session = session[UserSessionContext]
    except Exception:
        user_session = None

    if user_session:
        user_session.begin_using()
    try:
        yield
    finally:
        if user_session:
            user_session.end_using()
        CURRENT_SESSION = None


class ConnectionType(NamedConstant):
    DIRECT = lambda handler, *args: handler(args)
    QUEUED = lambda handler, *args: asyncio.get_running_loop().call_soon(handler, *args)


class SignalDecl:
    def __init__(self):
        self.handlers: dict[rio.Session, set[tuple[Callable[[...], None], ConnectionType]]] = defaultdict(set)

    def connect(self, handler: Callable[[...], None], type: ConnectionType = ConnectionType.QUEUED) -> bool:
        self.handlers[CURRENT_SESSION].add((handler, type))
        return True

    @staticmethod
    def _wrapper(ctx, handler: Callable[[...], None], *args):
        with ctx:
            handler(*args)

    def emit(self, *args) -> None:
        for handler, conn in self.handlers[CURRENT_SESSION]:
            conn.value(partial(self._wrapper, BTP.current(), handler), *args)


class MouseEvent(i.MouseEvent):
    __slots__ = ('event',)

    def __init__(self, event: rio.PointerEvent):
        self.event = event

    def is_left_button(self) -> bool:
        return self.event.button == 'left'

    def is_right_button(self) -> bool:
        return self.event.button == 'right'


class FontMetrics(i.FontMetrics):
    __slots__ = ('_widget',)

    def __init__(self, w: Widget):
        self._widget = w

    def average_char_width(self) -> int:
        return 1  # best guess -- rio measures sizes in char heights


class SizePolicy(Enum):
    MINIMUM_EXPANDING = ()
    PREFERRED = ()


class TEXT_ALIGN(NamedConstant):
    s_vertical = 0xF << 4

    LEFT = 1
    CENTER = 6
    RIGHT = 11
    TOP = LEFT << 4
    V_CENTER = CENTER << 4
    BOTTOM = RIGHT << 4

    @classmethod
    def from_str(cls, s: str) -> TEXT_ALIGN:
        return super().from_str(s.upper())  # type: ignore[return-value]

    def rio_attr(self) -> str:
        return 'align_y' if self.value & self.s_vertical else 'align_x'

    def rio_value(self) -> float:
        return ((self.value >> 4 if self.value & self.s_vertical else self.value) - 1) / 10


class DynamicComponent(rio.Component):
    builder: ComponentBuilder | None = None

    def __post_init__(self):
        self.key = f'dc_{id(self.builder)}'

    def build(self) -> rio.Component:
        assert self.builder, 'DynamicComponent has no builder'
        subcomponent = self.builder.build(self.session)
        if not self.builder.component:
            self.builder.component = self
            self.builder.subcomponent = subcomponent
        else:
            assert self.builder.component is self, 'DynamicComponent reused!'
        return subcomponent


class ComponentBuilder:
    __slots__ = ('_kwargs', 'component', 'subcomponent')

    s_component_class: type[rio.Component] = None
    s_forced_kwargs = {}
    s_default_kwargs = {}
    s_children_attr = 'children'
    s_single_child = False
    s_pass_children_in_kwargs = False
    s_size_adjustments = ('min_width', 'min_height', 'margin_left', 'margin_top', 'margin_right', 'margin_bottom', 'margin_x', 'margin_y', 'margin')
    s_layout_attrs = ('grow_x', 'grow_y', 'align_x', 'align_y')

    @staticmethod
    def current_session():
        return CURRENT_SESSION

    def _get_children(self) -> list:
        children = self.get_children()
        if self.s_single_child and children is None:
            return []
        return [children] if self.s_single_child else children

    def get_children(self):
        return self._kwargs[self.s_children_attr]

    def set_children(self, children):
        self._kwargs[self.s_children_attr] = children
        self.force_update()

    def child_count(self):
        return len(self.get_children())

    def _set_children(self, children):
        if self.s_single_child and not children:
            self.set_children(None)
        else:
            assert not self.s_single_child or len(children) == 1
            self.set_children(children[0] if self.s_single_child else children)

    def _make_kwargs(self, **kwargs):
        defaults = {kw: value(self, kwargs) if callable(value) else value for kw, value in self.s_default_kwargs.items()}
        return defaults | kwargs | self.s_forced_kwargs | dict(key=id(self))

    def __init__(self, *children, **kwargs):
        assert self.s_component_class, f'{self.__class__.__name__}: has no s_component_class'
        self._kwargs = self._make_kwargs(**kwargs)
        self.component = None
        self.subcomponent = None
        kw_kids = self._kwargs.get(self.s_children_attr)
        if self.s_single_child:
            self._set_children(children if children else [kw_kids])
        else:
            self._set_children([])
            self.add_children(*children)
            if kw_kids is not None:
                self.add_children(*kw_kids)

    def add_children(self, *children):
        existing_children = set(self._get_children())
        if self.s_single_child:
            new_children = [child for child in children if child is not None and child not in existing_children]
            if new_children:
                assert not existing_children
                self._set_children(new_children)
        else:
            self.get_children().extend(child for child in children if child is not None and child not in existing_children)
        self.force_update()

    def with_children(self, *args):
        if not args:
            return self
        return self.__class__(*args, **self._kwargs)

    def _build_children(self, session: rio.Session):
        return [child() if isinstance(child, ComponentBuilder) else child for child in self._get_children() if child is not None]

    @classmethod
    def create_component(cls, *children, **kwargs) -> rio.Component | None:
        if cls.s_component_class:
            if cls.s_pass_children_in_kwargs:
                # noinspection PyAugmentAssignment
                kwargs = kwargs | {'children': list(children)}
                children = ()
            return cls.s_component_class(*children, **kwargs)
        return None

    def build(self, session: rio.Session) -> rio.Component | None:
        kwargs = {k: v for k, v in self._kwargs.items() if k != self.s_children_attr}
        for size_adjustment in self.s_size_adjustments:
            if size_adjustment in kwargs:
                kwargs[size_adjustment] /= session.pixels_per_font_height
        children: list = self._build_children(session)
        return self.create_component(*children, **kwargs)

    def __call__(self) -> rio.Component:
        kw = {k: self[k] for k in self.s_layout_attrs if k in self}
        return DynamicComponent(builder=self, **kw) if not self.component else self.component

    def __getitem__(self, item):
        if hasattr(self.subcomponent, item):
            return getattr(self.subcomponent, item)
        return self._kwargs[item]

    def __contains__(self, item):
        return item in self._kwargs

    @classmethod
    def _not_supported(cls, message='not supported', item=None):
        item = item or inspect.stack()[1].function
        print(f'{cls.__name__}.{item}: - {message}')

    def __setitem__(self, item, value):
        if item != self.s_children_attr and not hasattr(self.s_component_class, item):
            self._not_supported(item=item)
            return

        if hasattr(self.subcomponent, item):
            setattr(self.subcomponent, item, value)
        self._kwargs[item] = value

    def force_update(self):
        component: rio.Component = self.component
        if component:
            component.force_refresh()

    def setdefault(self, item, default):
        try:
            value = self[item]
        except KeyError:
            value = self[item] = default
        return value

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def callback(self, callback):
        def cb(widget, *args, **kwargs):
            with session_context(widget.subcomponent.session):
                # note - callback must not yield the event loop!
                return callback(*args, **kwargs)

        return types.MethodType(cb, self)


class Widget(ComponentBuilder, i.Widget):
    __slots__ = ('_layout',)

    s_component_class = rio.Container
    s_stretch_arg = 'grow_x'
    s_default_layout_factory = lambda: FlowLayout()
    s_default_kwargs = dict(grow_y=False, align_y=0)
    s_unwrap_single_child = True

    def __init_subclass__(cls, **kwargs):
        cls.s_default_layout_factory = None
        cls.s_default_kwargs = super().s_default_kwargs | cls.s_default_kwargs

    def __init__(self, *children, **kwargs):
        super().__init__(*children, **kwargs)
        layout_factory = self.__class__.s_default_layout_factory
        self._layout = layout_factory() if layout_factory else None

    def __getstate__(self):
        assert getattr(self, '__dict__', self) is self, f'widgets should not have __dict__: {self}'

        def unbind(o):
            if isinstance(o, types.MethodType) and o.__self__:
                if isinstance((s := o.__self__), Traitable):
                    return '__rebind_traitable__', type(s), s.id(), o.__func__
                if o.__self__ is self:
                    return '__rebind_self__', o.__func__
                raise RuntimeError(f'cannot handle bound method {o}')
            return o

        def slot_set(c):
            slots = getattr(c, '__slots__', ())
            return {slots} if isinstance(slots, str) else set(slots)

        all_slots = reduce(operator.or_, (slot_set(base) for base in self.__class__.__mro__), set())
        return {k: unbind(getattr(self, k)) for k in all_slots}

    def __setstate__(self, state):
        for k, v in state.items():
            if isinstance(v, tuple):
                if v[0] == '__rebind_traitable__':
                    t_class, t_id, func = v[1:]
                    v = func.__get__(t_class(t_id))
                elif v[0] == '__rebind_self__':
                    v = v[1].__get__(self)
            setattr(self, k, v)

    def set_layout(self, layout: i.Layout):
        self._layout = layout

    def _build_children(self, session: rio.Session):
        return [self._layout.with_children(*self._get_children()).build(session)] if self._layout else super()._build_children(session)

    def set_stretch(self, stretch):
        assert stretch in (0, 1), 'Only stretch of 0 or 1 is currently supported'
        self[self.s_stretch_arg] = bool(stretch)

    def apply_style_sheet(self, style: dict) -> RC:
        if text_align := TEXT_ALIGN.from_str(style.pop('text-align', '')):
            self[text_align.rio_attr()] = text_align.rio_value()
        if hasattr(self.s_component_class, 'text_style'):
            ss = StyleSheet()
            rc = ss.set_values(sheet=style)
            self['text_style'] = ss.text_style or None
            return rc
        return StyleSheet.rc(style)

    def set_style_sheet(self, sh: str):
        from ui_10x.utils import UxStyleSheet  # TODO: circular import

        rc = self.apply_style_sheet(UxStyleSheet.loads(sh))
        if not rc:
            print(f'{self.__class__.__name__}.set_style_sheet: \n{rc.error()}')

    def set_minimum_height(self, height: int):
        self['min_height'] = height

    def set_minimum_width(self, width: int):
        self['min_width'] = width

    def set_size_policy(self, x_policy: SizePolicy, y_policy: SizePolicy):
        self['grow_x'] = x_policy == SizePolicy.MINIMUM_EXPANDING
        self['grow_y'] = y_policy == SizePolicy.MINIMUM_EXPANDING

    def set_tool_tip(self, tooltip: str):
        self['tooltip'] = tooltip

    def set_text(self, text: str):
        if self.s_single_child:
            self[self.s_children_attr] = text
            return
        self._not_supported()

    def set_read_only(self, read_only: bool):
        self['is_sensitive'] = not read_only

    def style_sheet(self) -> str:
        from ui_10x.utils import UxStyleSheet  # TODO: circular import

        if hasattr(self.s_component_class, 'text_style'):
            ss = StyleSheet()
            ss.text_style = self.get('text_style')
            return UxStyleSheet.dumps(ss.sheet)

        self._not_supported()
        return ''

    def set_enabled(self, enabled: bool):
        self['is_sensitive'] = enabled

    def font_metrics(self) -> FontMetrics:
        return FontMetrics(self)

    def set_geometry(self, *args):
        ##TODO
        self._not_supported()

    def width(self) -> int:
        ##TODO
        self._not_supported()
        return 0

    def height(self) -> int:
        ##TODO
        self._not_supported()
        return 0

    def map_to_global(self, point: Point) -> Point:
        ##TODO
        self._not_supported()
        return point

    @classmethod
    def create_component(cls, *children, **kwargs) -> rio.Component | None:
        if cls.s_unwrap_single_child and len(children) == 1 and isinstance(first_child := children[0], rio.Component):
            if kwargs:
                cls._not_supported(f'ignored kwargs {kwargs}')
            return first_child
        return super().create_component(*children, **kwargs)


class Layout(Widget, i.Layout):
    def add_widget(self, widget: Widget, stretch=None, **kwargs):
        assert not kwargs, f'kwargs not supported: {kwargs}'
        assert widget is not None
        if stretch is not None:
            widget.set_stretch(stretch)
        self.get_children().append(widget)

    def add_layout(self, layout: Layout, stretch=None, **kwargs):
        self.add_widget(layout, stretch=stretch, **kwargs)

    def set_spacing(self, spacing: int):
        assert spacing == 0, 'Only zero is supported'
        self['spacing'] = 0

    def set_contents_margins(self, left, top, right, bottom):
        self['margin_left'] = left
        self['margin_top'] = top
        self['margin_right'] = right
        self['margin_bottom'] = bottom


class FlowLayout(Layout, i.Layout):
    s_component_class = rio.FlowContainer


class Point(i.Point):
    __slots__ = ('_x', '_y')

    def __init__(self, x: int = 0, y: int = 0):
        self._x = x
        self._y = y

    def x(self) -> int:
        return self._x

    def y(self) -> int:
        return self._y



try:
    from dev_10x.version import __version__
except ImportError:
    __version__ = '0.0.0+unknown'  # fallback for dev envs

from py10x_infra import MongoCollectionHelper

__all__ = [ __version__, MongoCollectionHelper ]



from __future__ import annotations

from datetime import date

import pytest
from xx_common.rdate import (
    BIZDAY_ROLL_RULE,
    PROPAGATE_DATES,
    TENOR_FREQUENCY,
    RDate,
)


class TestRDate:
    """Unit tests for RDate class."""

    def test_init_with_symbol(self):
        """Test RDate constructor with symbol strings."""
        # Test basic symbols
        rd = RDate('3M')
        assert rd.freq == TENOR_FREQUENCY.MONTH
        assert rd.count == 3

        rd = RDate('1Y')
        assert rd.freq == TENOR_FREQUENCY.YEAR
        assert rd.count == 1

        rd = RDate('6Q')
        assert rd.freq == TENOR_FREQUENCY.QUARTER
        assert rd.count == 6

        rd = RDate('2W')
        assert rd.freq == TENOR_FREQUENCY.WEEK
        assert rd.count == 2

        rd = RDate('5C')
        assert rd.freq == TENOR_FREQUENCY.CALDAY
        assert rd.count == 5

        # Test negative values
        rd = RDate('-2M')
        assert rd.freq == TENOR_FREQUENCY.MONTH
        assert rd.count == -2

    def test_init_with_freq_and_count(self):
        """Test RDate constructor with frequency and count parameters."""
        rd = RDate(freq=TENOR_FREQUENCY.MONTH, count=3)
        assert rd.freq == TENOR_FREQUENCY.MONTH
        assert rd.count == 3

        # Test default count
        rd = RDate(freq=TENOR_FREQUENCY.YEAR)
        assert rd.freq == TENOR_FREQUENCY.YEAR
        assert rd.count == 1

    def test_init_invalid_symbol(self):
        """Test RDate constructor with invalid symbol raises AttributeError."""
        with pytest.raises(AttributeError, match="Invalid tenor symbol '3X'"):
            RDate('3X')

    def test_init_missing_freq(self):
        """Test RDate constructor without freq parameter raises AssertionError."""
        with pytest.raises(AssertionError, match='freq must be a valid TENOR_FREQUENCY'):
            RDate(count=3)

    def test_symbol(self):
        """Test symbol property returns correct string representation."""
        rd = RDate('3M')
        assert rd.symbol() == '3M'

        rd = RDate('1Y')
        assert rd.symbol() == '1Y'

        rd = RDate('-2Q')
        assert rd.symbol() == '-2Q'

    def test_to_str(self):
        """Test to_str method returns symbol."""
        rd = RDate('5C')
        assert rd.to_str() == '5C'

    def test_serialize_deserialize(self):
        """Test serialization and deserialization."""
        rd = RDate('3M')
        serialized = rd.serialize(embed=True)
        assert serialized == '3M'

        deserialized = RDate.deserialize(serialized)
        assert deserialized.freq == rd.freq
        assert deserialized.count == rd.count

    def test_from_str(self):
        """Test from_str class method."""
        rd = RDate.from_str('2Y')
        assert rd.freq == TENOR_FREQUENCY.YEAR
        assert rd.count == 2

    def test_from_any_xstr(self):
        """Test from_any_xstr class method."""
        # Test with RDate instance
        rd1 = RDate('3M')
        rd2 = RDate.from_any_xstr(rd1)
        assert rd2.freq == rd1.freq
        assert rd2.count == rd1.count

        # Test with tuple
        rd3 = RDate.from_any_xstr((TENOR_FREQUENCY.YEAR, 5))
        assert rd3.freq == TENOR_FREQUENCY.YEAR
        assert rd3.count == 5

        # Test invalid type
        with pytest.raises(AssertionError, match='unexpected type'):
            RDate.from_any_xstr('invalid')

    def test_same_values(self):
        """Test same_values class method."""
        rd1 = RDate('3M')
        rd2 = RDate('3M')
        rd3 = RDate('6M')

        assert RDate.same_values(rd1, rd2) is True
        assert RDate.same_values(rd1, rd3) is False

    def test_apply_basic(self):
        """Test apply method with basic calendar operations."""
        rd = RDate('3M')
        start_date = date(2023, 1, 15)

        # Create a mock calendar that treats all days as business days
        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

            def advance_bizdays(self, d, count):
                from datetime import timedelta

                return d + timedelta(days=count)

        cal = MockCalendar()

        # Test with NO_ROLL rule
        result = rd.apply(start_date, cal, BIZDAY_ROLL_RULE.NO_ROLL)
        expected = date(2023, 4, 15)  # 3 months later
        assert result == expected

    def test_apply_bizday(self):
        """Test apply method with business day frequency."""
        rd = RDate(freq=TENOR_FREQUENCY.BIZDAY, count=5)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

            def advance_bizdays(self, d, count):
                from datetime import timedelta

                return d + timedelta(days=count)

        cal = MockCalendar()

        start_date = date(2023, 1, 15)
        result = rd.apply(start_date, cal, BIZDAY_ROLL_RULE.NO_ROLL)
        expected = date(2023, 1, 20)  # 5 business days later
        assert result == expected

    def test_apply_rule(self):
        """Test apply_rule class method."""
        start_date = date(2023, 1, 15)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

            def advance_bizdays(self, d, count):
                from datetime import timedelta

                return d + timedelta(days=count)

        cal = MockCalendar()

        # Apply multiple tenors: 3M then 1Y
        result = RDate.apply_rule(start_date, cal, BIZDAY_ROLL_RULE.NO_ROLL, '3M,1Y')
        expected = date(2024, 4, 15)  # 3M from Jan 15 = Apr 15, then 1Y = Apr 15 2024
        assert result == expected

    def test_conversion_freq_multiplier(self):
        """Test conversion_freq_multiplier method."""
        # Same frequency
        rd1 = RDate('3M')
        rd2 = RDate('6M')
        assert rd1.conversion_freq_multiplier(rd2.freq) == 1.0

        # Different frequencies
        rd_year = RDate('1Y')
        rd_month = RDate('12M')
        assert rd_year.conversion_freq_multiplier(rd_month.freq) == 12

        rd_quarter = RDate('1Q')
        assert rd_year.conversion_freq_multiplier(rd_quarter.freq) == 4

        # Invalid conversions
        rd_bizday = RDate(freq=TENOR_FREQUENCY.BIZDAY, count=1)
        with pytest.raises(ValueError, match='cannot get a conversion multiplier'):
            rd_year.conversion_freq_multiplier(rd_bizday.freq)

    def test_equate_freq(self):
        """Test equate_freq method."""
        # Same frequency
        rd1 = RDate('3M')
        rd2 = RDate('6M')
        eq1, eq2 = rd1.equate_freq(rd2)
        assert eq1.freq == eq2.freq == TENOR_FREQUENCY.MONTH
        assert eq1.count == 3
        assert eq2.count == 6

        # Different frequencies - convert to common frequency
        rd_year = RDate('1Y')
        rd_month = RDate('12M')
        eq1, eq2 = rd_year.equate_freq(rd_month)
        assert eq1.freq == eq2.freq == TENOR_FREQUENCY.MONTH
        assert eq1.count == 12  # 1Y = 12M
        assert eq2.count == 12

    def test_multadd(self):
        """Test multadd method."""
        rd = RDate('6M')

        # Multiply by scalar
        result = rd.multadd(2.0, 0)
        assert result.freq == TENOR_FREQUENCY.MONTH
        assert result.count == 12

        # Add another RDate
        rd2 = RDate('3M')
        result = rd.multadd(1.0, rd2)
        assert result.freq == TENOR_FREQUENCY.MONTH
        assert result.count == 9

        # Invalid addition
        with pytest.raises(ValueError, match='cannot add a non-zero number'):
            rd.multadd(1.0, 5)

        # Invalid type
        with pytest.raises(ValueError, match='cannot calc a linear combination'):
            rd.multadd(1.0, [])

    def test_mathematical_operations(self):
        """Test mathematical operations (__mul__, __rmul__, __truediv__, __add__)."""
        rd = RDate('6M')

        # Multiplication
        result = rd * 2
        assert result.freq == TENOR_FREQUENCY.MONTH
        assert result.count == 12

        result = 3 * rd
        assert result.freq == TENOR_FREQUENCY.MONTH
        assert result.count == 18

        # Division by scalar
        result = rd / 2
        assert result.freq == TENOR_FREQUENCY.MONTH
        assert result.count == 3

        # Division by another RDate
        rd2 = RDate('3M')
        result = rd / rd2
        assert result == 2.0

        # Addition
        result = rd + rd2
        assert result.freq == TENOR_FREQUENCY.MONTH
        assert result.count == 9

    def test_from_tenors(self):
        """Test from_tenors class method."""
        tenors = RDate.from_tenors('3M,1Y,6Q')
        assert len(tenors) == 3
        assert tenors[0].symbol() == '3M'
        assert tenors[1].symbol() == '1Y'
        assert tenors[2].symbol() == '6Q'

        # Test with custom delimiter
        tenors = RDate.from_tenors('3M;1Y;6Q', delim=';')
        assert len(tenors) == 3

    def test_add_bizdays(self):
        """Test add_bizdays class method."""
        start_date = date(2023, 1, 15)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

            def advance_bizdays(self, d, count):
                from datetime import timedelta

                return d + timedelta(days=count)

        cal = MockCalendar()

        result = RDate.add_bizdays(start_date, 5, cal, BIZDAY_ROLL_RULE.NO_ROLL)
        expected = date(2023, 1, 20)
        assert result == expected

    def test_roll_to_bizday(self):
        """Test roll_to_bizday class method."""
        test_date = date(2023, 1, 15)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

        cal = MockCalendar()

        result = RDate.roll_to_bizday(test_date, cal, BIZDAY_ROLL_RULE.NO_ROLL)
        assert result == test_date  # NO_ROLL returns the date as-is

    def test_relop(self):
        """Test relop class method."""
        rd1 = RDate('3M')
        rd2 = RDate('6M')
        test_date = date(2023, 1, 15)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

        cal = MockCalendar()

        # Test less than
        result = RDate.relop(RDate.RELOP.LT, rd1, rd2, test_date, cal, BIZDAY_ROLL_RULE.NO_ROLL)
        # 3M from Jan 15 = Apr 15, 6M from Jan 15 = Jul 15, so Apr 15 < Jul 15
        assert result is True

        # Test equal
        rd3 = RDate('3M')
        result = RDate.relop(RDate.RELOP.EQ, rd1, rd3, test_date, cal, BIZDAY_ROLL_RULE.NO_ROLL)
        assert result is True

    def test_dates_schedule_forward(self):
        """Test dates_schedule method with forward propagation."""
        rd = RDate('3M')
        start_date = date(2023, 1, 15)
        end_date = date(2023, 7, 15)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

        cal = MockCalendar()

        dates = rd.dates_schedule(start_date, end_date, cal, BIZDAY_ROLL_RULE.NO_ROLL, PROPAGATE_DATES.FORWARD, allow_stub=True)

        # Should generate dates: Jan 15, Apr 15, Jul 15
        expected = [date(2023, 1, 15), date(2023, 4, 15), date(2023, 7, 15)]
        assert dates == expected

    def test_dates_schedule_backward(self):
        """Test dates_schedule method with backward propagation."""
        rd = RDate('3M')
        start_date = date(2023, 1, 15)
        end_date = date(2023, 7, 15)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

        cal = MockCalendar()

        dates = rd.dates_schedule(start_date, end_date, cal, BIZDAY_ROLL_RULE.NO_ROLL, PROPAGATE_DATES.BACKWARD, allow_stub=True)

        # Should generate dates: Jul 15, Apr 15, Jan 15 (backward from end)
        expected = [date(2023, 1, 15), date(2023, 4, 15), date(2023, 7, 15)]
        assert dates == expected

    def test_period_dates(self):
        """Test period_dates method."""
        rd = RDate('3M')
        start_date = date(2023, 1, 15)
        end_date = date(2023, 7, 15)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

        cal = MockCalendar()

        start_dates, end_dates, all_dates = rd.period_dates(
            start_date, end_date, cal, BIZDAY_ROLL_RULE.NO_ROLL, PROPAGATE_DATES.FORWARD, allow_stub=True
        )

        expected_all = [date(2023, 1, 15), date(2023, 4, 15), date(2023, 7, 15)]
        expected_starts = [date(2023, 1, 15), date(2023, 4, 15)]
        expected_ends = [date(2023, 4, 15), date(2023, 7, 15)]

        assert all_dates == expected_all
        assert start_dates == expected_starts
        assert end_dates == expected_ends

    def test_period_dates_for_tenor(self):
        """Test period_dates_for_tenor method."""
        rd = RDate('3M')  # frequency
        tenor = RDate('9M')  # total tenor
        start_date = date(2023, 1, 15)

        class MockCalendar:
            def is_bizday(self, d):
                return True

            def next_bizday(self, d):
                return d

            def prev_bizday(self, d):
                return d

        cal = MockCalendar()

        start_dates, end_dates, all_dates = rd.period_dates_for_tenor(
            start_date, tenor, cal, BIZDAY_ROLL_RULE.NO_ROLL, PROPAGATE_DATES.FORWARD, allow_stub=True
        )

        # Start at Jan 15, end at Oct 15 (9M later), with 3M periods
        expected_all = [date(2023, 1, 15), date(2023, 4, 15), date(2023, 7, 15), date(2023, 10, 15)]
        expected_starts = [date(2023, 1, 15), date(2023, 4, 15), date(2023, 7, 15)]
        expected_ends = [date(2023, 4, 15), date(2023, 7, 15), date(2023, 10, 15)]

        assert all_dates == expected_all
        assert start_dates == expected_starts
        assert end_dates == expected_ends
